#  legame/board_game.py
#
#  Copyright 2020 - 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
"""
Provides the BoardGame class, a framework for board games.
"""
import logging
from math import floor
from pygame.font import SysFont
from pygame.cursors import arrow, broken_x
from pygame import Rect, Surface, mouse
from pygame.draw import line
from pygame.sprite import Sprite
from pygame.math import Vector2 as Vector
from legame.game import Game, GameState
from legame.sprite_enhancement import MovingSprite


class BoardGame(Game):

	board					= None
	statusbar				= None

	my_color				= None
	opponent_color			= None

	def __init__(self, options = None):
		"""
		BoardGame constructor; calls Game.__init__() and instantiates the GameBoard and
		"""
		Game.__init__(self, options)
		self.board = self.get_board()
		self.statusbar = Statusbar()

	def initial_background(self, display_size):
		self.statusbar.rect.top = self.board.rect.height
		board_bg = self.board.initial_background(display_size)
		my_background = Surface((
			self.board.rect.width,
			self.board.rect.height + self.statusbar.rect.height
		))
		my_background.blit(board_bg, (0, 0))
		return my_background

	def get_board(self):
		return GameBoard()


class GameBoard:

	columns				= 7
	rows				= 9
	cell_width			= 50
	cell_height			= 50
	left				= 0
	top					= 0

	background_color	= (0,0,0)
	grid_lines_color	= (80,80,80)

	background			= None

	def __init__(self, columns = None, rows = None):
		if columns:
			self.columns = columns
		if rows:
			self.rows = rows
		self.rect = Rect((
			self.left, self.top,
			self.columns * self.cell_width,
			self.rows * self.cell_height + 1))
		self.__cells = [[None for y in range(self.rows)] for x in range(self.columns)]
		self.last_column = self.columns - 1
		self.last_row = self.rows - 1
		self.center_column = self.columns // 2
		self.center_row = self.rows // 2
		self.cell_half_width = self.cell_width // 2
		self.cell_half_height = self.cell_height // 2

	def initial_background(self, display_size):
		bg = Surface(self.rect.size)
		bg.fill(self.background_color)
		for col in range(self.columns):
			x = col * self.cell_width
			line(bg, self.grid_lines_color, (x, 0), (x, self.rect.height), 1)
		for row in range(self.rows + 1):
			y = row * self.cell_height
			line(bg, self.grid_lines_color, (0, y), (self.rect.width, y), 1)
		return bg

	# Set / get / inspect board positions:

	def cell_at(self, *args):
		"""
		Return a Cell from screen coordinates.
		Args may be a pair of numbers (float or int), or a tuple of numbers (float or int).
		Returns None if the coordinates are outside of the board.
		"""
		if len(args) == 1:
			x, y = args[0]
		elif len(args) == 2:
			x, y = args
		else:
			raise ValueError("Board.cell_at() takes two numbers or a tuple of two numbers")
		return Cell(
			floor((x - Game.current.board.left) / Game.current.board.cell_width),
			floor((y - Game.current.board.top) / Game.current.board.cell_height)
		) if Game.current.board.rect.collidepoint(int(x), int(y)) else None

	def piece_at(self, cell):
		"""
		Returns a reference to the GamePiece occupying the given cell.
		If no piece occupies the cell, returns None.
		"""
		assert isinstance(cell, Cell)
		return self.__cells[cell.column][cell.row]

	def set_cell(self, cell, piece):
		"""
		Places a reference to the given GamePiece in the given cell.
		"""
		assert isinstance(cell, Cell)
		assert isinstance(piece, AbstractGamePiece)
		self.__cells[cell.column][cell.row] = piece
		return self

	def clear_cell(self, cell):
		"""
		Kills the GamePiece at the given cell, if one exists there.
		"""
		assert isinstance(cell, Cell)
		self.__cells[cell.column][cell.row] = None
		return self

	def is_mine(self, cell):
		"""
		Returns True if there is a GamePiece at the given cell which has this player's
		"color".
		"""
		assert isinstance(cell, Cell)
		piece = self.piece_at(cell)
		return False if piece is None else piece.color == Game.current.my_color

	def is_opponents(self, cell):
		"""
		Returns True if there is a GamePiece at the given cell, and it doesn't have
		this player's "color".
		"""
		assert isinstance(cell, Cell)
		piece = self.piece_at(cell)
		return False if piece is None else piece.color != Game.current.my_color

	def is_empty(self, cell):
		"""
		Returns True if the given cell is empty.
		"""
		assert isinstance(cell, Cell)
		return self.piece_at(cell) is None

	def column(self, column):
		"""
		Returns a list of the occupants of the given column.
		"""
		return self.__cells[column]

	def row(self, row):
		"""
		Returns a list of the occupants of the given row.
		"""
		return [ self.__cells[column][row] for column in range(self.columns) ]

	def rotate(self, cell):
		"""
		Returns a board position rotated 180 degrees.
		Used for showing opponent moves when the move is defined from their perspective.
		"""
		assert isinstance(cell, Cell)
		return Cell(self.last_column - cell.column, self.last_row - cell.row)

	def dump(self):
		print("  " + "".join([f"{column:2d}" for column in range(self.columns)]))
		print("\n".join(
			[f"{row:2d}|" + "|".join(self.__cells[column][row].color \
				if self.__cells[column][row] is not None else " " \
			for column in range(self.columns)) + "|" for row in range(self.rows)]
		))
		print("  " + "-" * (self.columns * 2 + 1))


class Cell:

	def __init__(self, column, row):
		self.column, self.row = column, row

	def __iter__(self):
		return iter((self.column, self.row))

	def __eq__(self, cell):
		return self.column == cell.column and self.row == cell.row

	def set(self, piece):
		"""
		Places a reference to the given GamePiece in this cell.
		"""
		Game.current.board.set_cell(self, piece)
		return self

	def clear(self):
		"""
		Kills the GamePiece at this cell, if one exists there.
		"""
		Game.current.board.clear_cell(self)
		return self

	def piece(self):
		"""
		Returns the content of the GameBoard at this cell
		"""
		return Game.current.board.piece_at(self)

	def is_mine(self):
		"""
		Returns True if there is a GamePiece at this cell which has this player's "color".
		"""
		return Game.current.board.is_mine(self)

	def is_opponents(self):
		"""
		Returns True if there is a GamePiece at this cell, and it doesn't have this
		player's "color".
		"""
		return Game.current.board.is_opponents(self)

	def is_empty(self):
		"""
		Returns True if this cell is empty.
		"""
		return Game.current.board.is_empty(self)

	def center(self):
		"""
		Returns the center point of this position.
		"""
		return (
			Game.current.board.left + Game.current.board.cell_width * self.column \
				+ Game.current.board.cell_half_width,
			Game.current.board.top + Game.current.board.cell_height * self.row \
				+ Game.current.board.cell_half_height
		)

	def rect(self):
		"""
		Returns a pygame rect which covers this position.
		Top-left is the top-left of the cell.
		"""
		return Rect(
			Game.current.board.left + Game.current.board.cell_width * self.column,
			Game.current.board.top + Game.current.board.cell_height * self.row,
			Game.current.board.cell_width,
			Game.current.board.cell_height
		)

	def copy(self):
		"""
		Returns a copy.
		"""
		return Cell(self.column, self.row)

	def shifted(self, columns = None, rows = None):
		"""
		Returns a copy of this Cell shifted by the given "columns" and "rows".
		Either of those arguments are optional. Providing neither will yield a copy.
		"""
		return Cell(
			self.column if columns is None else self.column + columns,
			self.row if rows is None else self.row + rows
		)

	def moved(self, column = None, row = None):
		"""
		Returns a copy of this Cell with either "columns" or "rows" set to the given
		values(s), and the other values unchanged.
		Either of those arguments are optional. Providing neither will yield a copy.
		"""
		return Cell(
			self.column if column is None else column,
			self.row if row is None else row
		)

	def __str__(self):
		return f"Cell at column {self.column}, row {self.row}"


class Statusbar(Sprite):

	padding				= 8
	font				= 'Ubuntu Light'
	font_size			= 22
	foreground_color	= (250,250,160)
	background_color	= (0,0,12)

	def __init__(self):
		Sprite.__init__(self, Game.current.sprites)
		Game.current.sprites.change_layer(self, Game.LAYER_ABOVE_BG)
		self.font = SysFont(self.font, 22)
		self.image = Surface((
			Game.current.board.rect.width,
			self.font.get_linesize() + self.padding * 2
		))
		self.text = ""
		self.rect = self.image.get_rect()

	def clear(self):
		self.image.fill(self.background_color)

	def write(self, text, color = None):
		self.text = text
		if color is not None:
			self.foreground_color = color
		self._update()

	def append(self, text):
		self.text = self.text + text
		self._update()

	def _update(self):
		self.clear()
		self.image.blit(
			self.font.render(self.text, True, self.foreground_color),
			(self.padding, self.padding)
		)


class BoardGameState(GameState):

	mouse_down_pos	= None
	may_click		= False

	def __init__(self, **kwargs):
		"""
		BoardGameState constructor - sets the current mouse position passes execution
		to GameState constructor.
		"""
		self.mouse_pos = Game.current.board.cell_at(mouse.get_pos())
		GameState.__init__(self, **kwargs)

	def loop_end(self):
		"""
		Cyclical function called at the end of Game._main_loop()
		Sets the mouse cursor based on the state of the "may_click" flag.
		"""
		if self.may_click:
			mouse.set_cursor(*arrow)
		else:
			mouse.set_cursor(*broken_x)

	def mouse_motion(self, event):
		"""
		Mouse move event passed to this GameState.
		event will contain:	pos, rel, buttons
		"""
		cell = Game.current.board.cell_at(event.pos)
		if cell is None:
			return
		if self.mouse_pos is not None \
			and (cell.column != self.mouse_pos.column or cell.row != self.mouse_pos.row):
			self.mouse_exit(self.mouse_pos)
			self.mouse_enter(cell)
		self.mouse_pos = cell

	def mouse_button_down(self, event):
		"""
		Mouse down event passed to this GameState.
		event will contain:	pos, button
		"""
		self.mouse_down_pos = self.mouse_pos

	def mouse_button_up(self, event):
		"""
		Mouse up event passed to this GameState.
		event will contain: pos, button
		"""
		if self.may_click and self.mouse_down_pos == self.mouse_pos:
			self.click(self.mouse_pos, event)

	def mouse_enter(self, cell):
		"""
		"Pseudo" event which occurs after the mouse moved to a new position on the board.
		This event immediately follows "mouse_exit".
		"""

	def mouse_exit(self, cell):
		"""
		"Pseudo" event which occurs after the mouse moves out of a position on the board.
		This event is immediately followed by "mouse_enter" with the new position given.
		"""

	def click(self, cell, event):
		"""
		"Pseudo" event which occurs when the player presses and releases the mouse
		button over a single position.
		"cell" is the board position which was "clicked" (not screen_rect x/y).
		"event" is the pygame event passed to the "mousebuttonup" function, which
		will contain "pos" and "button" attributes.
		"""


class AbstractGamePiece:
	"""
	An "abstract" version of a GamePiece, used for testing and assertions
	"""

	def __init__(self, cell, color):
		self.cell = cell
		self.color = color
		self.rect = self.cell.rect()
		self.position = Vector(self.cell.center())
		self.target_cell = None
		self._motion_function = None
		Game.current.board.set_cell(cell, self)

	def move_to(self, *,
		target_cell = None,
		column = None, row = None,
		columns = None, rows = None,
		on_arrival = None):
		"""
		High-level command which sets this GamePiece on a path towards a given cell.
		Each subsequent cyclic update will move this one frame closer to "target_cell".
		When travel is complete, the optional "on_arrival" function will be called.

		You can specify a target using one of the following three options:

		1. Pass a Cell as "target_cell"
		2. Specify the absolute target "column" and/or "row".
		3. Specify the relative "columns" and/or "rows" to move by.

		In the cases 2 and 3 above, if any of "column", "columns", "row" or "rows" is not given,
		this GamePiece's current column or row will be used.

		"""
		def arrival_function():
			self.cell = self.target_cell
			current_resident = Game.current.board.piece_at(self.cell)
			if current_resident is not None:
				current_resident.kill()
			Game.current.board.set_cell(self.cell, self)
			self._motion_function = self.no_motion
			if callable(on_arrival):
				on_arrival()
		if target_cell is None:
			self.target_cell = self.cell.copy()
			self.target_cell.column = column if column is not None \
				else self.cell.column + columns if columns is not None \
				else self.cell.column
			self.target_cell.row = row if row is not None \
				else self.cell.row + rows if rows is not None \
				else self.cell.row
		else:
			self.target_cell = target_cell
		if self.target_cell == self.cell:
			logging.warn("GamePiece.move_to target cell is the current cell")
		else:
			Game.current.board.clear_cell(self.cell)
			return self.travel_to(self.target_cell.center(), arrival_function)

	def travel_to(self, coords, on_arrival):
		"""
		High-level command which sets a MovingSprite on a path towards a given target.
		Each subsequent call to "move()" will move it one frame closer to the target.
		When travel is complete, the optional "on_arrival" function will be called.
		"""


class GamePiece(MovingSprite, Sprite, AbstractGamePiece):

	max_speed	= 17.0
	min_speed	= 0.5
	decel_rate	= 1.4

	def __init__(self, cell, color):
		AbstractGamePiece.__init__(self, cell, color)
		self._motion_function = self.cartesian_motion
		self.motion = Vector()
		self.image_set = Game.current.resources.image_set(f"{self.__class__.__name__}/{self.color}")
		Sprite.__init__(self, Game.current.sprites)
		Game.current.sprites.change_layer(self, Game.LAYER_PLAYER)


#  end legame/board_game.py



#  legame/examples/alien-herd.py
#
#  Copyright 2020 - 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
"""
Visual variant of "herd.py" (which demonstrates "neighbor" detection using the
Neighborhood class.)
"""
import argparse, logging
from random import seed, randrange
from legame.examples.herd import HerdDemo, Forager, Predator, GSWatch


class AlienHerdDemo(HerdDemo):

	cells_x			= 16
	cells_y			= 11
	bg_color		= (0,0,14)
	num_foragers	= 120
	num_predators	= 10

	def initial_state(self):
		for i in range(50):
			AienForager(randrange(20, 210), randrange(20, 110), randrange(0, 360))
		for i in range(7):
			AlienPredator(randrange(140, 400), randrange(100, 300), randrange(0, 360))
		return GSWatch()


class AienForager(Forager):
	color				= (60,120,25)
	running_color		= (90,200,45)
	low_speed			= 0.18
	high_speed			= 1.125
	turn_sluggish		= 30
	accel_sluggish		= 44


class AlienPredator(Predator):
	color				= (160,22,16)
	running_color		= (180,52,28)
	eat_color			= (248,20,40)
	low_speed			= 0.3
	high_speed			= 0.95
	turn_sluggish		= 10
	accel_sluggish		= 20


if __name__ == '__main__':
	p = argparse.ArgumentParser()
	p.add_argument("--verbose", "-v", action = "store_true",
		help = "Show more detailed debug information")
	p.epilog = __doc__
	options = p.parse_args()
	logging.basicConfig(
		level = logging.DEBUG if options.verbose else logging.ERROR,
		format = "[%(filename)24s:%(lineno)-4d] %(message)s"
	)

	seed()
	p.exit(AlienHerdDemo().run())


#  end legame/examples/alien-herd.py



#  legame/sprite_enhancement.py
#
#  Copyright 2020 - 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
"""
Provides classes for easier positioning and moving of sprites.
"""
import math
from pygame import Rect
from pygame.math import Vector2 as Vector
from legame import	triangular, turning_degrees, \
					OFFSCREEN_LEFT, OFFSCREEN_TOP, OFFSCREEN_RIGHT, OFFSCREEN_BOTTOM, \
					COMPASS_WEST, COMPASS_NORTH, COMPASS_EAST, COMPASS_SOUTH, \
					SIDE_LEFT, SIDE_TOP, SIDE_RIGHT, SIDE_BOTTOM


def to_vector(target):
	"""
	Returns a Vector from the given tuple, Vector, or CenteredSprite.
	Raises ValueError if impossible to convert.
	"""
	if isinstance(target, CenteredSprite):
		return target.position
	if isinstance(target, Vector):
		return target
	if isinstance(target, tuple):
		return Vector(target)
	raise ValueError("Invalid target: %s" % target)


class CenteredSprite:

	height				= 10
	width				= 10

	def __init__(self, x = 0.0, y = 0.0):
		"""
		CenteredSprite constructor.
		Creates a "rect" property as used by pygame.sprite.Sprite, with the center
		at the x/y position of this Sprite.
		"x" and "y" must be float values which define the position of the object.
		"""
		self.position = Vector(x, y)
		self.rect = Rect(
			int(self.position.x - self.width / 2),
			int(self.position.y - self.height / 2),
			self.width,
			self.height
		)

	@property
	def x(self):
		"""
		Gets the "x" value of this object's positiong vector, relative to the Surface on which this
		object is to be rendered.
		"""
		return self.position.x

	@x.setter
	def x(self, value):
		"""
		Sets the "x" value of this object's positiong vector, relative to the Surface on which this
		object is to be rendered.
		"""
		self.position.x = value

	@property
	def y(self):
		"""
		Gets the "y" value of this object's positiong vector, relative to the Surface on which this
		object is to be rendered.
		"""
		return self.position.y

	@y.setter
	def y(self, value):
		"""
		Sets the "y" value of this object's positiong vector, relative to the Surface on which this
		object is to be rendered.
		"""
		self.position.y = value

	def update_rect(self):
		"""
		Updates the "rect" used by the Sprite class with this CenteredSprite's "position" vector.
		"""
		self.rect.centerx = int(self.position.x)
		self.rect.centery = int(self.position.y)
		return self


class MovingSprite(CenteredSprite):
	"""
	A class compatible with pygame.Sprite which tracks position, motion, rotation in 2d space.
	"""

	max_speed			= 100.0	# Absolute speed limit for this thing
	min_speed			= 0.5	# ...and the slowest its allowed to go
	accel_rate			= 1.0	# Value added to speed each frame when accelerating
	decel_rate			= 1.0	# Value subtracted from speed each frame when decelerating
	turning_speed		= 0.0	# Current turning speed (rotation in degrees per frame)
	max_turning_speed	= 360	# Max degrees per frame this can turn; 360 = unlimited
	destination			= None  # Necessary for seek_motion
	_motion_function	= None	# Function called to move this on Sprite.update()
	_arrival_function	= None	# Function called when destination reached by seeking motion

	def __init__(self, x = 0.0, y = 0.0, speed = None, direction = None):
		"""
		MovingSprite constructor.
		"x" and "y" must be float values which define the position of the object.
		"speed" must be a float value which sets the magnitude of the motion vector of the object.
		"direction" must be a float value which sets the motion direction of the object in degrees.

		If you would like to use a motion vector made up of x/y values, create a
		"motion" vector with the values you need, and skip this function completely,
		like so:

			class MySprite(MovingSprite):
				def __init__(self):
					CenteredSprite.__init__(self, <x>, <y>)			# Sets Sprite.rect
					self.motion = Vector(<x>, <y>)					# Cartesian
					self._motion_function = self.cartesian_motion	# (example)

		"""
		CenteredSprite.__init__(self, x, y)
		self._motion_function = self.cartesian_motion
		self.motion = Vector()
		if speed is None or direction is None: return
		self.motion.from_polar((speed, direction))

	@property
	def speed(self):
		"""
		Get the "magnitude" of this object's motion vector.
		"""
		return self.motion.magnitude()

	@speed.setter
	def speed(self, value):
		"""
		Sets the "magnitude" of this object's motion vector.
		"""
		try:
			self.motion.scale_to_length(value)
		except ValueError as e:
			self.motion.from_polar((value, 0.0))

	@property
	def direction(self):
		"""
		Gets the screen direction of this MovingSprite in degrees
		"""
		mag, deg = self.motion.as_polar()
		return deg

	@direction.setter
	def direction(self, degrees):
		"""
		Sets the screen direction of this MovingSprite to the given degrees.
		"""
		mag = self.motion.magnitude()
		if mag == 0.0:
			self.motion.from_polar((2.22e-222, degrees))
		else:
			self.motion.from_polar((mag, degrees))

	def update(self):
		"""
		Regular cyclic update task, as called from pygame.Sprite.
		Calls the current move function, which by default is "cartesian_motion".
		"""
		self._motion_function()

	def shift_position(self, x, y):
		"""
		Does a relative cartesian move the position of the object by x, y.
		"""
		self.position.x += x
		self.position.y += y
		self.rect.centerx = int(self.position.x)
		self.rect.centery = int(self.position.y)
		return self

	def set_motion_polar(self, magnitude, degrees):
		"""
		Set the current motion vector from the given magnitude, degrees.
		"""
		self.motion.from_polar((magnitude, degrees))

	def travel_to(self, target, on_arrival = None):
		"""
		High-level command which sets this MovingSprite on a path towards a given target.
		Each subsequent call to "move()" will move it one frame closer to the target.
		When travel is complete, the optional "on_arrival" function will be called.
		"""
		self.destination = to_vector(target)
		self._motion_function = self.seek_motion
		self._arrival_function = on_arrival
		return self.make_heading()

	def make_heading(self):
		"""
		Sets the initial speed and direction of an object provided a destination
		using the "travel_to()" function.

		The default implementation sets this object's direction to the direction to its
		destination, without taking into consideration turning speed. The speed of this
		MovingSprite is set to it's acceration rate, unless it's already moving faster.

		A true physical moving thing most likely cannot change speed and direction in
		such an immediate fashion, unless the things you hear about "off world" vehicles
		visiting our planet are really true. You're free to model that, I suppose.

		This implementation is sufficient for game pieces such as used in BoardGame,
		but should be overridden in order to create more sophisticated motion.
		"""
		speed = max(self.speed, self.accel_rate)
		self.motion = (self.destination - self.position)
		self.motion.scale_to_length(speed)
		return self

	####################################################################################
	# Move routines which may be called in the "update()" function of the parent Sprite:
	# Various movement methods, called from the "update" method, depending upon the
	# instance and circumstances. The "update" method calls the "_motion_function"
	# which is a reference to one of these methods, or a method in a derived class.
	####################################################################################

	def no_motion(self):
		"""
		A "_motion_function" which does nothing.
		"""

	def cartesian_motion(self):
		"""
		Moves one frame by adding the x/y values of this MovingSprite's "motion" vector to the
		x/y values of this MovingSprite's "position" vector.
		This is the default "_motion_function".
		"""
		self.position.x += self.motion.x
		self.position.y += self.motion.y
		return self.update_rect()

	def seek_motion(self):
		"""
		One of the "_motion_function" methods, called from "update". This method points
		the sprite towards its current "destination" vector and accelerates towards it,
		decelerating to zero when it approaches.

		This is the move function set in the "travel_to" method.

		This implementation is sufficient for game pieces as used in BoardGame,
		but should be overridden in order to create more sophisticated motion.
		"""
		remaining_distance = (self.destination - self.position).magnitude()
		if remaining_distance <= self.decel_rate:
			self.position = self.destination
			self.rect.centerx = int(self.position.x)
			self.rect.centery = int(self.position.y)
			self.speed = 0.0
			self._motion_function = self.no_motion
			if self._arrival_function:
				self._arrival_function()
		elif remaining_distance <= triangular(self.speed, self.decel_rate):
			# coming up on target_pos; decelerate:
			self.speed = max(self.min_speed, self.speed - self.decel_rate)
		else:
			# still far away from target_pos; accelate towards it:
			self.speed = min(self.max_speed, self.speed + self.accel_rate)
		return self.cartesian_motion()

	####################################################################################
	# High-level movement functions
	####################################################################################

	def turn_towards(self, vector):
		"""
		Turn incrementally towards the given position vector, taking into account this
		MovingSprite's current and max turning speed.
		"""
		self.increment_turning_speed(self.direction_to_vector(vector))
		self.direction += self.turning_speed

	def turn_away_from(self, vector):
		"""
		Turn incrementally away from the given position vector, taking into account
		this MovingSprite's current and max turning speed.
		"""
		self.increment_turning_speed(self.direction_to_vector(vector) + 180)
		self.direction += self.turning_speed

	def increment_turning_speed(self, direction):
		"""
		Incrementally adjust this MovingSprite's "turning_speed" attribute towards or
		away from the given screen direction in degrees, taking into account this
		MovingSprite's current and max turning speed.

		Affects only the "turning_speed" attribute - does NOT change the "motion" vector.
		"""
		desired_turn = turning_degrees(self.direction - direction)
		turn_mag = abs(desired_turn)
		if turn_mag < self.max_turning_speed:
			self.direction = direction
			self.turning_speed = 0.0
		elif abs(self.turning_speed) < self.max_turning_speed:
			self.turning_speed -= math.copysign(self.max_turning_speed, desired_turn)

	####################################################################################
	# Relative position methods:
	####################################################################################

	def distance_to_vector(self, vector):
		"""
		Returns a float - the distance in pixels from this MovingSprite to the given
		position vector.
		"""
		return self.position.distance_to(vector)

	def distance_to_thing(self, thing):
		"""
		Returns a float - the distance in pixels from this MovingSprite to the given
		"thing-like" object.
		"""
		return self.position.distance_to(thing.position)

	def direction_to_vector(self, vector):
		"""
		Returns the world angle in degrees from this MovingSprite's position to the
		given position vector. Returns non-normalized degrees (float).

		If you need a Vector instead of an angle in radians, use Vector subtraction.
		"""
		return (vector - self.position).as_polar()[1]

	def direction_to_thing(self, thing):
		"""
		Returns the world angle in degrees from this MovingSprite's position to the
		given "thing-like" object's position. Returns non-normalized degrees (float).

		If you need a Vector instead of an angle in radians, use Vector subtraction.
		"""
		return (thing.position - self.position).as_polar()[1]

	def angle_to_vector(self, vector):
		"""
		Returns the world angle in degrees from this MovingSprite's position to the
		given position vector. Returns non-normalized degrees (float).

		If you need a Vector instead of an angle in radians, use Vector subtraction.

		Alias of "direction_to_vector".
		"""
		return (vector - self.position).as_polar()[1]

	def angle_to_thing(self, thing):
		"""
		Returns the world angle in degrees from this MovingSprite's position to the
		given "thing-like" object's position. Returns non-normalized degrees (float).

		If you need a Vector instead of an angle in radians, use Vector subtraction.

		Alias of "direction_to_thing".
		"""
		return (thing.position - self.position).as_polar()[1]

	def radians_to_vector(self, vector):
		"""
		Returns the world angle in radians from this MovingSprite's position to the
		given position vector. Returns non-normalized radians (float).

		If you need a Vector, simply subtract this MovingSprite's position from the vector's.
		"""
		return math.radians((vector - self.position).as_polar()[1])

	def radians_to_thing(self, thing):
		"""
		Returns the world angle in radians from this MovingSprite's position to the
		given "thing-like" object's position. Returns non-normalized radians (float).

		If you need a Vector, simply subtract this MovingSprite's position from the vector's.
		"""
		return math.radians((thing.position - self.position).as_polar()[1])

	####################################################################################
	# Status methods:
	####################################################################################

	def is_offscreen(self, screen_rect):
		"""
		Returns one of the "OFFSCREEN_<direction>" constants if this MovingSprite's position
		if outside of the boundaries of the Surface it is to be rendered on.
		"""
		if self.rect.colliderect(screen_rect):
			bits = 0
			if self.position.x < 0:
				bits |= OFFSCREEN_LEFT
			elif self.position.x > screen_rect.width:
				bits |= OFFSCREEN_RIGHT
			if self.position.y < 0:
				bits |= OFFSCREEN_TOP
			elif self.position.y > screen_rect.height:
				bits |= OFFSCREEN_BOTTOM
			return bits
		return 0

	def is_directionless(self):
		"""
		Returns boolean "True" if radians and magnitude of this MovingSprite's motion Vector
		cannot be determined.
		"""
		return self.motion.x == 0 and self.motion.y == 0

	def ns_direction(self):
		"""
		Returns COMPASS_NORTH or COMPASS_SOUTH, depending on if the y-axis of movement is positive or negative
		"""
		return COMPASS_NORTH if self.motion.y < 0 else COMPASS_SOUTH

	def ew_direction(self):
		"""
		Returns COMPASS_EAST or COMPASS_WEST, depending on if the y-axis of movement is positive or negative
		"""
		return COMPASS_WEST if self.motion.x < 0 else COMPASS_EAST

	def __str__(self):
		if self.is_directionless():
			return "<Motionless %s at %.1f / %.1f>" % (type(self).__name__, self.position.x, self.position.y)
		return "<%s at %.1f / %.1f, moving %d degrees at %.1f pixels-per-second>" % \
			(type(self).__name__, self.position.x, self.position.y, self.motion.degrees, self.motion.magnitude())


class BoxedInSprite:
	"""
	A class which can add boundary checking to a MovingSprite.
	"""

	def __init__ (self, boundary):
		"""
		Initialize boundary checking.
		"boundary" is a pygame Rect within which movement is to be constrained.
		Tip: Rect.inflate() can be used to reduce the size of a Rect with a margin around the outside.
		So if you want to restrict movement of a 40px x 40px sprite, such that the outer edges of the
		sprite never leave the screen, you can create a boundary Rect that is 20px (half of the width
		/ height) smaller than the screen Rect using:

			(Screen Rect).inflate(-20)

		"""
		self.boundary = boundary

	def nearest_boundary(self):
		"""
		Determine if near one of the boundary walls.
		Returns None if not near.
		Returns a tuple of (<distance from>, <which wall>) if outside boundary.
		<distance from> is nearest distance to the wall, not distance along MovingSprite's direction
		<which wall> is one of the SIDE_<side> constants defined in locals
		"""
		if self.x < self.boundary.left:
			# Near the west wall
			return (self.x, SIDE_LEFT)
		if self.y < self.boundary.top:
			# Near the north wall
			return (self.y, SIDE_TOP)
		if self.x > self.boundary.right:
			# Near the east wall
			return (self.boundary.right - self.x, SIDE_RIGHT)
		if self.y > self.boundary.bottom:
			# Near the south wall
			return (self.boundary.bottom - self.y, SIDE_BOTTOM)
		return None

	def update(self):
		tup = self.nearest_boundary()
		if tup is not None:
			if tup[1] == SIDE_TOP or tup[1] == SIDE_BOTTOM:
				self.motion.y = - self.motion.y
			else:
				self.motion.x = - self.motion.x
		self.cartesian_motion()


#  end legame/sprite_enhancement.py



#  legame/configurable.py
#
#  Copyright 2020 - 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
import pickle
from os import mkdir
from os.path import dirname, isdir, isfile, join
from appdirs import user_config_dir


class Configurable:

	config_file			= ""
	config				= {}

	def load_config(self):
		config_dir = user_config_dir(self.__class__.__name__)
		self.config_file = join(config_dir, "settings.dat")
		if isfile(self.config_file):
			with open(self.config_file, "rb") as fh:
				self.config = pickle.load(fh)

	def save_config(self):
		if self.config:
			config_dir = dirname(self.config_file)
			if not isdir(config_dir):
				mkdir(config_dir)
			with open(self.config_file, "wb") as fh:
				pickle.dump(self.config, fh)


#  end legame/configurable.py



#  legame/tests/neighbors_test.py
#
#  Copyright 2020 - 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
from pygame import Rect
from legame.sprite_enhancement import MovingSprite
from legame.neighbors import Neighborhood, Neighbor, Quadrant


class Thing(MovingSprite, Neighbor):
	pass
	def __init__(self, area, x, y):
		MovingSprite.__init__(self, x, y)
		area.observe(self)
	def notice(self, neighbor):
		assert isinstance(neighbor, Neighbor)


def test_constructor():
	rect = Rect(0, 0, 400, 300)

	# Calc height/width from cell count correctly
	nh = Neighborhood(rect, 4, 3)
	assert nh.cell_width == 100
	assert nh.cell_height == 100
	assert len(nh.all_quadrants) == (nh.cells_x - 1) * (nh.cells_y - 1)
	assert len(nh.cells) == nh.cells_x
	assert len(nh.cells[0]) == nh.cells_y

	# Check "quadrants" populated correctly:
	for quadrant in nh.all_quadrants:
		assert isinstance(quadrant, Quadrant)

	# Check "cells" populated correctly:
	assert len(nh.cells) == nh.cells_x
	assert len(nh.cells[0]) == nh.cells_y

	# top-left cell (x = 0, y = 0):
	cell = nh.cells[0][0]
	assert isinstance(cell, list)
	assert len(cell) == 1
	assert isinstance(cell[0], Quadrant)
	assert cell[0].x == 0
	assert cell[0].y == 0

	# top, second from the left (x = 1, y = 0):
	cell = nh.cells[1][0]
	assert isinstance(cell, list)
	assert len(cell) == 2
	assert isinstance(cell[0], Quadrant)
	assert cell[0].x == 0
	assert cell[0].y == 0
	assert isinstance(cell[1], Quadrant)
	assert cell[1].x == 1
	assert cell[1].y == 0

	# top, third from the left (x = 2, y = 0):
	cell = nh.cells[2][0]
	assert isinstance(cell, list)
	assert len(cell) == 2
	assert isinstance(cell[0], Quadrant)
	assert cell[0].x == 1
	assert cell[0].y == 0
	assert isinstance(cell[1], Quadrant)
	assert cell[1].x == 2
	assert cell[1].y == 0

	# top-right cell (x = 3, y = 0):
	cell = nh.cells[3][0]
	assert isinstance(cell, list)
	assert len(cell) == 1
	assert isinstance(cell[0], Quadrant)
	assert cell[0].x == 2
	assert cell[0].y == 0

	# middle-left cell (x = 0, y = 1):
	cell = nh.cells[0][1]
	assert isinstance(cell, list)
	assert len(cell) == 2
	assert isinstance(cell[0], Quadrant)
	assert cell[0].x == 0
	assert cell[0].y == 0
	assert isinstance(cell[1], Quadrant)
	assert cell[1].x == 0
	assert cell[1].y == 1

	# middle row, second from the left (x = 1, y = 1):
	cell = nh.cells[1][1]
	assert isinstance(cell, list)
	assert len(cell) == 4
	assert isinstance(cell[0], Quadrant)
	assert cell[0].x == 0
	assert cell[0].y == 0
	assert isinstance(cell[1], Quadrant)
	assert cell[1].x == 0
	assert cell[1].y == 1
	assert isinstance(cell[2], Quadrant)
	assert cell[2].x == 1
	assert cell[2].y == 0
	assert isinstance(cell[3], Quadrant)
	assert cell[3].x == 1
	assert cell[3].y == 1

	# middle row, third from the left (x = 2, y = 1):
	cell = nh.cells[2][1]
	assert isinstance(cell, list)
	assert len(cell) == 4
	assert isinstance(cell[0], Quadrant)
	assert cell[0].x == 1
	assert cell[0].y == 0
	assert isinstance(cell[1], Quadrant)
	assert cell[1].x == 1
	assert cell[1].y == 1
	assert isinstance(cell[2], Quadrant)
	assert cell[2].x == 2
	assert cell[2].y == 0
	assert isinstance(cell[3], Quadrant)
	assert cell[3].x == 2
	assert cell[3].y == 1

	# middle row, right (x = 3, y = 1):
	cell = nh.cells[3][1]
	assert isinstance(cell, list)
	assert len(cell) == 2
	assert isinstance(cell[0], Quadrant)
	assert cell[0].x == 2
	assert cell[0].y == 0
	assert isinstance(cell[1], Quadrant)
	assert cell[1].x == 2
	assert cell[1].y == 1

	# bottom-left cell (x = 0, y = 2):
	cell = nh.cells[0][2]
	assert isinstance(cell, list)
	assert len(cell) == 1
	assert isinstance(cell[0], Quadrant)
	assert cell[0].x == 0
	assert cell[0].y == 1

	# bottom row, second from the left (x = 1, y = 2):
	cell = nh.cells[1][2]
	assert isinstance(cell, list)
	assert len(cell) == 2
	assert isinstance(cell[0], Quadrant)
	assert cell[0].x == 0
	assert cell[0].y == 1
	assert isinstance(cell[1], Quadrant)
	assert cell[1].x == 1
	assert cell[1].y == 1

	# bottom row, third from the left (x = 2, y = 2):
	cell = nh.cells[2][2]
	assert isinstance(cell, list)
	assert len(cell) == 2
	assert isinstance(cell[0], Quadrant)
	assert cell[0].x == 1
	assert cell[0].y == 1
	assert isinstance(cell[1], Quadrant)
	assert cell[1].x == 2
	assert cell[1].y == 1

	# bottom-right (x = 3, y = 2):
	cell = nh.cells[3][2]
	assert isinstance(cell, list)
	assert len(cell) == 1
	assert isinstance(cell[0], Quadrant)
	assert cell[0].x == 2
	assert cell[0].y == 1

def test_sprite_population():
	rect = Rect(0, 0, 40, 30)
	nh = Neighborhood(rect, 4, 3)

	t1 = Thing(nh, 5, 5)
	assert len(nh._observed_sprites_list) == 1
	assert len(nh.quadrant(0,0).sprites) == 0
	assert len(nh.sprites_in(0,0)) == 0
	assert len(nh.quadrant(0,0).sprites) == 0
	assert len(nh.sprites_in(0,0)) == 0

	nh.notify_sprites()
	assert len(nh.quadrant(0,0).sprites) == 1
	assert len(nh.sprites_in(0,0)) == 1

	t2 = Thing(nh, 15, 15)
	nh.notify_sprites()
	assert len(nh.quadrant(0,0).sprites) == 2
	assert len(nh.sprites_in(0,0)) == 2
	assert len(nh.quadrant(1,1).sprites) == 1
	assert len(nh.sprites_in(1,1)) == 1

	t3 = Thing(nh, 25, 25)
	nh.notify_sprites()
	assert len(nh.quadrant(0,0).sprites) == 2
	assert len(nh.sprites_in(0,0)) == 2
	assert len(nh.quadrant(1,1).sprites) == 2
	assert len(nh.sprites_in(1,1)) == 2
	assert len(nh.quadrant(2,1).sprites) == 1
	assert len(nh.sprites_in(2,1)) == 1

	t4 = Thing(nh, 35, 15)
	nh.notify_sprites()
	assert len(nh.quadrant(0,0).sprites) == 2
	assert len(nh.sprites_in(0,0)) == 2
	assert len(nh.quadrant(1,1).sprites) == 2
	assert len(nh.sprites_in(1,1)) == 2
	assert len(nh.quadrant(2,1).sprites) == 2
	assert len(nh.sprites_in(2,1)) == 2


#  end legame/tests/neighbors_test.py



#  legame/examples/network-game.py
#
#  Copyright 2020 - 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
"""
Demonstrates sending moves over a network.
"""
import argparse, logging
import random
try:
	from pygame.locals import K_q, K_ESCAPE
except ImportError:
	from pygame import K_q, K_ESCAPE
from pygame.sprite import Sprite
from legame.game import Game, GameState, GameStateFinal
from legame.board_game import BoardGame, GameBoard, GamePiece, Cell
from legame.flipper import Flipper, FlipBetween, FlipThrough, FlipNone
from legame.network_game import NetworkGame
from legame.exit_states import ExitAnimation


class TestGame(BoardGame, NetworkGame):

	xfer_interval	= 0.1		# Number of seconds between calls to service the messenger

	def __init__(self, options = None):
		self.set_resource_dir_from_file(__file__)
		BoardGame.__init__(self, options)
		NetworkGame.__init__(self, options)

	def get_board(self):
		return GameBoard(5, 5)

	def initial_state(self):
		global board, statusbar, send, me, my_opponent
		board = Game.current.board
		statusbar = Game.current.statusbar
		send = Game.current.messenger.send
		return GSWhoGoesFirst()

# ----------------------------------------
# Game states:

class GSBase(GameState):
	"""
	Used as the base class of all game states defined in this module.
	"""

	def key_down(self, event):
		"""
		Exit game immediately if K_ESCAPE or "q" key pressed
		"""
		if event.key in (K_ESCAPE, K_q):
			GSQuit(who = "me")


class GSWhoGoesFirst(GSBase):

	may_click	= False

	def enter_state(self):
		logging.debug("Enter state GSWhoGoesFirst")
		self.pick_a_color()

	def pick_a_color(self):
		logging.debug("pick_a_color()")
		self.my_roll = MsgPickAColor()
		self.my_roll.color = random.choice(["r", "g", "b"])
		self.my_roll.number = random.randrange(1, 255)
		send(self.my_roll)
		statusbar.write("Rolling for %s (my number is %d)" % (self.my_roll.color, self.my_roll.number))

	def handle_message(self, message):
		if isinstance(message, MsgPickAColor):
			if message.number == 0:
				if message.color == self.my_roll.color:
					raise Exception("Other machine conceded same color")
				Game.current.opponent_color = message.color
				logging.debug("Received concession: %s" % message.color)
				GSMyMove()
			elif message.number == self.my_roll.number:
				logging.debug("Picked the same number - trying again")
				self.pick_a_color()
			elif message.color == self.my_roll.color:
				# Both picked the same color. The one with the higher number gets their pick:
				if self.my_roll.number > message.number:
					logging.debug("Got dibs: %s; waiting for concession" % self.my_roll.color)
					Game.current.my_color = self.my_roll.color
					# Wait for concession MsgPickAColor with number = 0
				else:
					# Send concession MsgPickAColor with number = 0
					Game.current.opponent_color = message.color
					colors = ["r", "g", "b"]
					colors.remove(message.color)
					Game.current.my_color = random.choice(colors)
					self.my_roll = MsgPickAColor()
					self.my_roll.color = Game.current.my_color
					self.my_roll.number = 0
					logging.debug("Senging concession: %s" % self.my_roll.color)
					send(self.my_roll)
					GSWaitYourTurn()
			else:
				# Picked different colors. The one with the higher number moves first:
				Game.current.my_color = self.my_roll.color
				Game.current.opponent_color = message.color
				if message.number > self.my_roll.number:
					logging.debug("Picked different colors - I move first")
					GSMyMove()
				else:
					logging.debug("Picked different colors - they move first")
					GSWaitYourTurn()
		elif isinstance(message, MsgQuit):
			GSQuit(who = "them")
		else:
			logging.error("Unexpected message during GSWhoGoesFirst: %s" % message)
			GSQuit(who = "me")


class GSMyMove(GSBase):

	def enter_state(self):
		statusbar.write("GSMyMove")
		for y in range(board.rows):
			for x in range(board.columns):
				cell = Cell(x, y)
				if board.piece_at(cell) is None:
					Block(cell, Game.current.my_color)
					send(MsgAdd(cell = cell))
					GSWaitYourTurn()
					return
		GSQuit(who = "me")


class GSWaitYourTurn(GSBase):

	def enter_state(self):
		statusbar.write("GSWaitYourTurn")

	def handle_message(self, message):
		if isinstance(message, MsgAdd):
			message.rotate()
			board.set_cell(message.cell, Block(message.cell, Game.current.opponent_color))
		elif isinstance(message, MsgQuit):
			GSQuit(who = "them")
		else:
			logging.error("Unexpected message during GSWaitYourTurn: %s" % message)
			GSQuit(who = "me")
		GSMyMove()


class GSQuit(GameStateFinal):

	def enter_state(self):
		if self.who == "me": send(MsgQuit())
		ExitAnimation("bye-bye.png")

# ----------------------------------------
# Game pieces and other sprites:

class Block(GamePiece, Flipper):

	def __init__(self, cell, color):
		self.image_folder = "Block/" + color
		GamePiece.__init__(self, cell, color)
		Flipper.__init__(self, FlipThrough("enter", fps = 25), FlipNone())
		self.__glow = None

	def update(self):
		GamePiece.update(self)
		Flipper.update(self)

	def jiggle(self):
		self.flip(FlipBetween("jiggle", loops = 11, fps = 30), FlipNone())
		return self

	def glow(self):
		self.__glow = Glow(self.cell)
		return self

	def unglow(self):
		if self.__glow:
			self.__glow.kill()
			self.__glow = None
		return self


class Glow(Flipper, Sprite):

	def __init__(self, cell, frame = 0):
		self.cell = cell
		Sprite.__init__(self, Game.current.sprites)
		Game.current.sprites.change_layer(self, Game.LAYER_BELOW_PLAYER)
		Flipper.__init__(self, FlipBetween(loop_forever = True, frame = frame, fps = 30))
		self.rect = self.cell.get_rect()


if __name__ == '__main__':
	p = argparse.ArgumentParser()
	p.add_argument("--transport", type = str, default = "json")
	p.add_argument("--quiet", "-q", action = "store_true", help = "Don't make sound")
	p.add_argument("--verbose", "-v", action = "store_true", help = "Show more detailed debug information")
	p.add_argument("--direct", "-d", action = "store_true", help = "Connect by ip address instead of using udp broadcast discovery.")
	p.epilog = __doc__
	options = p.parse_args()
	logging.basicConfig(
		level = logging.DEBUG if options.verbose else logging.ERROR,
		format = "[%(filename)24s:%(lineno)-4d] %(message)s"
	)

	if options.transport == "json":
		from cable_car.json_messages import Message, MsgQuit

		class MsgAdd(Message):

			def encoded_attributes(self):
				return { "cell" : (self.cell.column, self.cell.row) }

			def decode_attributes(self, attributes):
				self.cell = Cell(*attributes["cell"])

			def rotate(self):
				self.cell = board.rotate(self.cell)
				return self

		class MsgPickAColor(Message):
			pass

	else:
		from cable_car.byte_messages import Message, MsgQuit

		class MsgAdd(Message):
			code = 0x8

			def encode(self):
				"""
				Encode column, row:
				"""
				return bytearray([self.cell.column, self.cell.row])

			def decode(self, msg_data):
				"""
				Read column, row from message data.
				"""
				self.cell = Cell(msg_data[0], msg_data[1])

			def rotate(self):
				self.cell = board.rotate(self.cell)
				return self

		class MsgPickAColor(Message):
			code = 0x9

			def encode(self):
				"""
				Encode color, number
				"""
				return bytearray([self.number]) + self.color.encode("ASCII")

			def decode(self, msg_data):
				"""
				Read color and number from message data.
				"""
				self.number, self.color = msg_data[0], msg_data[1:].decode()

	p.exit(TestGame(options).run())


#  end legame/examples/network-game.py



#  legame/examples/herd.py
#
#  Copyright 2020 - 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
"""
Demonstrates "neighbor" detection using the Neighborhood class
"""
import argparse, logging
from random import seed, randrange, uniform
from pygame.sprite import Sprite
from pygame import Rect, Surface
from pygame.draw import circle
try:
	from pygame.locals import SRCALPHA, K_q, K_ESCAPE
except ImportError:
	from pygame import SRCALPHA, K_q, K_ESCAPE
from legame.game import Game, GameState
from legame.sprite_enhancement import MovingSprite, BoxedInSprite
from legame.neighbors import Neighborhood, Neighbor
from legame import	SIDE_LEFT, SIDE_TOP, SIDE_RIGHT, SIDE_BOTTOM, \
					DEGREES_SOUTHEAST, DEGREES_NORTHEAST, DEGREES_NORTHWEST, DEGREES_SOUTHWEST


class HerdDemo(Game, Neighborhood):

	cells_x			= 12
	cells_y			= 8
	cell_width		= 60
	cell_height		= 60
	bg_color		= (240,230,140)
	num_foragers	= 60
	num_predators	= 7

	def __init__ (self):
		global game, neighborhood
		self.set_resource_dir_from_file(__file__)
		self.quiet = True
		Game.__init__(self)
		self.rect = Rect(0, 0, self.cell_width * self.cells_x, self.cell_width * self.cells_y)
		neighborhood = Neighborhood(self.rect.inflate(-30, -30), self.cells_x, self.cells_y)
		game = self

	def initial_background(self, display_size):
		bg = Surface(self.rect.size)
		bg.fill(self.bg_color)
		return bg

	def initial_state(self):
		r = neighborhood.rect
		midleft, midtop = neighborhood.rect.center
		for i in range(self.num_foragers):
			Forager(
				randrange(r.left, midleft),
				randrange(r.top, midtop),
				randrange(0, 360)
			)
		for i in range(self.num_predators):
			Predator(
				randrange(midleft, r.right),
				randrange(midtop, r.bottom),
				randrange(0, 360)
			)
		return GSWatch()

	def _end_loop(self):
		neighborhood.notify_sprites()
		self._state.loop_end()


class GSWatch(GameState):

	def key_down(self, event):
		"""
		Exit game immediately if K_ESCAPE key pressed
		"""
		if event.key in (K_ESCAPE, K_q):
			game.shutdown()

	def quit_event(self, event):
		"""
		Event handler called when the user clicks the window's close button.
		event will be empty
		"""
		game.shutdown()


class Animal(Neighbor, BoxedInSprite, MovingSprite, Sprite):

	height				= 8
	width				= 8

	def __init__(self, x, y, direction):
		Sprite.__init__(self, game.sprites)
		MovingSprite.__init__(self, x, y, self.low_speed, direction)
		BoxedInSprite.__init__(self, game.screen_rect.inflate(-20, -20))
		neighborhood.observe(self)
		self._normal_image = self.make_image(self.color)
		self._bright_image = self.make_image(self.running_color)
		self._accel = (self.high_speed - self.low_speed) / self.accel_sluggish
		self._decel = self._accel * 0.27
		self.max_turning_speed = 90 / self.turn_sluggish
		self._turn_min = -self.max_turning_speed
		self._turn_max = self.max_turning_speed
		self._nearest_animal = None
		self._nearest_animal_distance = None

	def make_image(self, color):
		surf = Surface((8,8), SRCALPHA)
		circle(surf, color, (4, 4), 4)
		return surf

	def boundary_check(self):
		"""
		Calls MovingSprite.cartesian_motion after boundary checking
		"""
		tup = self.nearest_boundary()
		if tup is not None:
			if tup[1] == SIDE_TOP:
				self.direction = DEGREES_SOUTHEAST if self.motion.x > 0 else DEGREES_SOUTHWEST
			elif tup[1] == SIDE_BOTTOM:
				self.direction = DEGREES_NORTHEAST if self.motion.x > 0 else DEGREES_NORTHWEST
			if tup[1] == SIDE_RIGHT:
				self.direction = DEGREES_SOUTHWEST if self.motion.y > 0 else DEGREES_NORTHWEST
			elif tup[1] == SIDE_LEFT:
				self.direction = DEGREES_SOUTHEAST if self.motion.y > 0 else DEGREES_NORTHEAST
			self.turning_speed = 0
		self.cartesian_motion()

	def kill(self):
		neighborhood.ignore(self)
		Sprite.kill(self)


class Forager(Animal):

	color				= (60,120,25)
	running_color		= (90,200,45)
	low_speed			= 0.25
	high_speed			= 1.33
	turn_sluggish		= 24
	accel_sluggish		= 44

	def notice(self, neighbor):
		if isinstance(neighbor, Predator):
			d = self.distance_to_thing(neighbor)
			if self._nearest_animal is None or d < self._nearest_animal_distance:
				self._nearest_animal = neighbor
				self._nearest_animal_distance = d

	def update(self):
		"""
		Default "update" function.
		"""
		if self._nearest_animal is None:
			self.image = self._normal_image
			self.speed = max(self.low_speed, self.speed - self._decel)
			self.turning_speed = uniform(self._turn_min, self._turn_max)
		else:
			self.image = self._bright_image
			self.turn_away_from(self._nearest_animal.position)
			self.speed = min(self.high_speed, self.speed + self._accel)
		self.boundary_check()
		self.cartesian_motion()
		self._nearest_animal = None


class Predator(Animal):

	color				= (140,42,16)
	running_color		= (180,72,28)
	eat_color			= (248,20,40)
	low_speed			= 0.3
	high_speed			= 0.88
	turn_sluggish		= 6
	accel_sluggish		= 14

	max_vigor			= 600
	tire_limit			= 50
	rested_limit		= 400
	eat_frames			= 240

	def __init__(self, x, y, direction):
		Animal.__init__(self, x, y, direction)
		self._eat_image = self.make_image(self.eat_color)
		self._vigor = 400
		self._tired = False
		self.update = self.hunting

	def notice(self, neighbor):
		if isinstance(neighbor, Forager):
			d = self.distance_to_thing(neighbor)
			if self._nearest_animal is None or d < self._nearest_animal_distance:
				self._nearest_animal = neighbor
				self._nearest_animal_distance = d

	def cartesian_motion(self):
		self._nearest_animal = None
		super().cartesian_motion()

	def hunting(self):
		if self._nearest_animal and self._nearest_animal.alive():
			self.image = self._bright_image
			if self._nearest_animal_distance < 4:
				#print("hunting -> kill_prey -> catching  vigor: %d" % self._vigor)
				self.kill_prey()
			else:
				self.image = self._bright_image
				self.turn_towards(self._nearest_animal.position)
				self.speed = min(self.high_speed, self.speed + self._accel)
				self._vigor -= 1
				#print("hunting -> chasing  vigor: %d" % self._vigor)
				self.update = self.chasing
		else:
			self.image = self._normal_image
			self.speed = max(self.low_speed, self.speed - self._decel)
			self.turning_speed = uniform(self._turn_min, self._turn_max)
			self.boundary_check()
		self.cartesian_motion()

	def chasing(self):
		self.image = self._bright_image
		if self._nearest_animal and self._nearest_animal.alive():
			if self._nearest_animal_distance < 4:
				#print("chasing -> kill_prey -> catching  vigor: %d" % self._vigor)
				self.kill_prey()
			elif self._vigor < self.tire_limit:
				self.turning_speed = 0.0
				#print("chasing -> going_to_rest  vigor: %d" % self._vigor)
				self.update = self.going_to_rest
			else:
				self.turn_towards(self._nearest_animal.position)
				self.speed = min(self.high_speed, self.speed + self._accel)
				self._vigor -= 1
		else:
			#print("chasing -> hunting  vigor: %d" % self._vigor)
			self.update = self.hunting
		self.cartesian_motion()

	def kill_prey(self):
		self._nearest_animal.kill()
		self.turning_speed = 0.0
		self.speed = max(0.0, self.speed - self._decel)
		self.update = self.catching

	def catching(self):
		self.image = self._eat_image
		self.speed = max(0.0, self.speed - self._decel)
		if self.speed == 0.0:
			#print("catching -> eating  vigor: %d" % self._vigor)
			self.update = self.eating
			self.eat_frames_countdown = self.eat_frames
		self.cartesian_motion()

	def eating(self):
		self._vigor = max(self.max_vigor, self._vigor + 2)
		self.eat_frames_countdown -= 1
		if self.eat_frames_countdown <= 0:
			#print("eating -> hunting  vigor: %d" % self._vigor)
			self.update = self.hunting

	def going_to_rest(self):
		self.speed -= self._decel
		self.cartesian_motion()
		if self.speed <= self._decel:
			self.speed = 0.0
			#print("going_to_rest -> resting  vigor: %d" % self._vigor)
			self.update = self.resting

	def resting(self):
		self._vigor += 1
		if self._vigor > self.rested_limit:
			#print("resting -> hunting  vigor: %d" % self._vigor)
			self.update = self.hunting


if __name__ == '__main__':
	p = argparse.ArgumentParser()
	p.add_argument("--verbose", "-v", action = "store_true", help = "Show more detailed debug information")
	p.epilog = __doc__
	options = p.parse_args()
	logging.basicConfig(
		level = logging.DEBUG if options.verbose else logging.ERROR,
		format = "[%(filename)24s:%(lineno)-4d] %(message)s"
	)

	seed()
	p.exit(HerdDemo().run())


#  legame/examples/herd.py



#  legame/game.py
#
#  Copyright 2020 - 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
"""
Provides the Game and GameState classes, a framework for writing games.
"""
import os, logging
import pygame
try:
	from pygame.locals import FULLSCREEN, SCALED, DOUBLEBUF, USEREVENT, NUMEVENTS
except ImportError:
	from pygame import FULLSCREEN, DOUBLEBUF, USEREVENT, NUMEVENTS
	SCALED = 0
from pygame.event import event_name
from pygame import Surface
from legame.resources import Resources


class Game:
	"""
	This is the base class of games created with the legame library.

	Game.current is a class variable which will contain a reference to the instance
	of the game currently running, and can be used as a "global" variable thoughout
	the code.
	"""

	current				= None		# Not-too-clever substitute for a global variable

	# Game options:
	quiet				= False		# Inhibit sound; mixer is not initialized
	fullscreen			= False
	resource_dump		= False		# Resources debugging mode; load filenames instead of files

	# resource manager settings:
	resource_dir		= None		# Directory where game data is located,
									# should contain subfolders "images" and "sounds"
									# See "set_resource_dir_from_file(__file__)"
	# display settings:
	fps 				= 60
	display_flags		= DOUBLEBUF
	display_depth		= 32
	caption				= ""
	icon				= None

	# sound settings:
	mixer_frequency		= 44100
	mixer_bitsize		= -16
	mixer_channels		= 8
	mixer_buffer		= 512

	# game objects:
	screen_rect			= None
	background			= None
	screen				= None
	sprites				= None
	resources			= None

	# timer options:
	max_timers			= 8

	# internal state management:
	_state				= None		# It's pretty important to keep this managed, hence, it's "protected"
	_stay_in_loop		= True		# Setting this to "False" exits the game, calling "exit_loop()"
	_next_state			= None		# Next game state waiting for change at end of main loop

	LAYER_BG			= 1			#
	LAYER_ABOVE_BG		= 2			#
	LAYER_BELOW_PLAYER	= 4			# Sprite layers
	LAYER_PLAYER		= 6			#
	LAYER_ABOVE_PLAYER	= 8			#
	LAYER_OVERLAY		= 10		#

	def __init__(self, options = None):
		"""

		Set options, start the necessary pygame modules, instantiate Resources, create
		the "sprites" group.

		This function passes the "resource_dir" attribute when instantiating a
		Resources instance to allow it to find images and sounds. You must set this
		attribute before calling this function. If you have a "resources" directory
		beneath the directory containing your game, the helper function
		"set_resource_dir_from_file" can set it for you quite easily.

		The "options" argument is expected to be a dictionary, the items of which are
		set as attributes of the game during initialization. Some appropriate key/value
		pairs to pass to the __init__ function would be:

			display_depth
			fps
			fullscreen
			quiet
			resource_dump

		A typical use case would be if you used the "argparse" library to read
		command-line options, and wish to pass those options to the Game class before
		starting up modules. i.e.:

			import argparse
			p = argparse.ArgumentParser()
			p.add_argument("--quiet", "-q", action = "store_true", help = "Don't make sound")
			p.add_argument("--fullscreen", "-f", action = "store_true", help = "Show fullscreen")
			options = p.parse_args()
			# Setup logging, etc...
			.
			.
			p.exit(MyGameClass(options).run())

		"""
		Game.current = self
		if options is not None:
			for varname, value in options.__dict__.items():
				setattr(self, varname, value)
		if self.resource_dir is None:
			self.resource_dir = "resources"
		self.resources = Resources(self.resource_dir, self.resource_dump)
		if not self.quiet:
			pygame.mixer.pre_init(self.mixer_frequency, self.mixer_bitsize,
			self.mixer_channels, self.mixer_buffer)
		pygame.init()
		self.sprites = pygame.sprite.LayeredUpdates()

		# Event handler mapping.
		self._event_handlers = dict([
			(getattr(pygame.locals, const), getattr(self, '_evt_' + const.lower())) \
			for const in dir(pygame.locals) \
			if hasattr(self, '_evt_' + const.lower())])

		# Fill-in the remaining timer events:
		for event_type in range(USEREVENT, NUMEVENTS + 1):
			self._event_handlers[event_type] = self.__timer_event
		self.__timer_callbacks = [None for x in range(self.max_timers)]
		self.__timer_arguments = [None for x in range(self.max_timers)]
		self.__timer_recur_flag = [None for x in range(self.max_timers)]

	def set_resource_dir_from_file(self, filename):
		"""
		Set the resource directory to a subfolder of the given file's parent folder.
		If your main game file is located at:

			"/home/user/some/path/game.py"

		... this function will return:

			"/home/user/some/path/resources"

		... which is a decent place to put your "images" and "sounds" folders.
							("Use the defaults, Luke")
		"""
		self.resource_dir = os.path.join(os.path.dirname(os.path.realpath(filename)), "resources")
		logging.debug('set resource dir to %s', self.resource_dir)

	def run(self):
		"""
		Run the game. The "initial_state" function is called from here, and must return
		an object of class GameState.
		"""
		self.show()
		self._state = self.initial_state()
		self._state.enter_state()	# Immediately enter state, not waiting for "_main_loop()"
		self._next_state = None	# Clear this, as it was set in "change_state()"
		self._main_loop()
		pygame.quit()
		return 0

	def show(self):
		"""
		Initilizes the screen.

		Make sure that you implement "initial_background", as that is what will be shown.

		It is safe to call this function multiple times.
		"""
		display_size = pygame.display.Info().current_w, pygame.display.Info().current_h
		self.background = self.initial_background(display_size)
		self.screen_rect = self.background.get_rect()
		if self.fullscreen:
			self.display_flags |= (FULLSCREEN | SCALED)
		else:
			os.environ['SDL_VIDEO_CENTERED'] = '1'
		if self.icon is not None:
			pygame.display.set_icon(pygame.image.load(os.path.join(
				self.resources.image_folder, self.icon)))
		pygame.display.set_caption(self.caption)
		self.screen = pygame.display.set_mode(
			self.screen_rect.size,
			self.display_flags,
			pygame.display.mode_ok(self.screen_rect.size, self.display_flags, self.display_depth)
		)
		self.screen.blit(self.background, (0,0))
		pygame.display.flip()

	def initial_background(self, display_size):
		"""
		Returns a pygame.Surface to use to fill the screen at startup.
		"""
		return Surface(display_size)

	def initial_state(self):
		"""
		Returns a GameState object which will take over as soon as the game is initialized.
		"""
		return GameState()

	def shutdown(self):
		"""
		Triggers the _main_loop to exit. The _main_loop will finish its current
		iteration before doing so.
		"""
		self._stay_in_loop = False

	def change_state(self, game_state):
		"""
		Lines up the next game state. The next time through main_loop, the new state
		will be current.

		It is possible that this function will be called more than once during a single
		game loop cycle. For example, a player might make a move and transition to a
		game state which waits for the opponent to move, at almost the same time as
		receiving a message from their opponent that they left the game.

		In such circumstances, the last call to this function takes precedence - with
		one caveat. If any game state subclasses GameStateFinal, the game state may not
		be changed at all.
		"""
		if isinstance(self._next_state, GameStateFinal):
			logging.warning("Cannot change game state when current state is GameStateFinal")
		else:
			self._next_state = game_state

	###############################################################################################

	def _main_loop(self):
		clock = pygame.time.Clock()
		while self._stay_in_loop:
			self._state.loop_start()
			for event in pygame.event.get():
				try:
					self._event_handlers[event.type](event)
				except KeyError:
					logging.warning('Unknown event "%s"', event_name(event.type))

			self._end_loop()
			self.sprites.update()
			self.sprites.clear(self.screen, self.background)
			pygame.display.update(self.sprites.draw(self.screen))
			if self._next_state:
				self._state.exit_state(self._next_state)
				self._state = self._next_state
				self._next_state = None
				self._state.enter_state()
			clock.tick(self.fps)
		for cls in self.__class__.mro():
			if "exit_loop" in cls.__dict__:
				cls.exit_loop(self)

	###############################################################################################

	def _end_loop(self):
		"""
		Called at the end of the _main_loop().
		The default implementation is to call "loop_end()" on the current GameState.
		This behaviour is overridden in NetworkGame in order to handle message transfer.
		"""
		self._state.loop_end()

	def exit_loop(self):
		"""
		Called when _main_loop() exits, after the final round of moving sprites and
		updating the display.
		"""

	# Event handlers:

	def _evt_activeevent(self, event):
		self._state.active_event(event)

	def _evt_audiodeviceadded(self, event):
		self._state.audio_device_added(event)

	def _evt_audiodeviceremoved(self, event):
		self._state.audio_device_removed(event)

	def _evt_controlleraxismotion(self, event):
		self._state.controller_axis_motion(event)

	def _evt_controllerbuttondown(self, event):
		self._state.controller_button_down(event)

	def _evt_controllerbuttonup(self, event):
		self._state.controller_button_up(event)

	def _evt_controllerdeviceadded(self, event):
		self._state.controller_device_added(event)

	def _evt_controllerdeviceremapped(self, event):
		self._state.controller_device_remapped(event)

	def _evt_controllerdeviceremoved(self, event):
		self._state.controller_device_removed(event)

	def _evt_dropbegin(self, event):
		self._state.drop_begin_event(event)

	def _evt_dropcomplete(self, event):
		self._state.drop_complete_event(event)

	def _evt_dropfile(self, event):
		self._state.drop_file_event(event)

	def _evt_droptext(self, event):
		self._state.drop_text_event(event)

	def _evt_fingerdown(self, event):
		self._state.finger_down(event)

	def _evt_fingermotion(self, event):
		self._state.finger_motion(event)

	def _evt_fingerup(self, event):
		self._state.finger_up(event)

	def _evt_joyaxismotion(self, event):
		self._state.joy_axis_motion(event)

	def _evt_joyballmotion(self, event):
		self._state.joy_ball_motion(event)

	def _evt_joybuttondown(self, event):
		self._state.joy_button_down(event)

	def _evt_joybuttonup(self, event):
		self._state.joy_button_up(event)

	def _evt_joydeviceadded(self, event):
		self._state.joy_device_added(event)

	def _evt_joydeviceremoved(self, event):
		self._state.joy_device_removed(event)

	def _evt_joyhatmotion(self, event):
		self._state.joy_hat_motion(event)

	def _evt_keydown(self, event):
		self._state.key_down(event)

	def _evt_keyup(self, event):
		self._state.key_up(event)

	def _evt_midiin(self, event):
		self._state.midi_in(event)

	def _evt_midiout(self, event):
		self._state.midi_out(event)

	def _evt_mousebuttondown(self, event):
		self._state.mouse_button_down(event)

	def _evt_mousebuttonup(self, event):
		self._state.mouse_button_up(event)

	def _evt_mousemotion(self, event):
		self._state.mouse_motion(event)

	def _evt_mousewheel(self, event):
		self._state.mouse_wheel(event)

	def _evt_multigesture(self, event):
		self._state.multi_gesture_event(event)

	def _evt_quit(self, event):
		self._state.quit_event(event)

	def _evt_syswmevent(self, event):
		self._state.sys_wm_event(event)

	def _evt_textediting(self, event):
		self._state.text_editing_event(event)

	def _evt_textinput(self, event):
		self._state.text_input_event(event)

	def _evt_videoexpose(self, event):
		self._state.video_expose(event)

	def _evt_videoresize(self, event):
		self._state.video_resize(event)

	def _evt_windowclose(self, event):
		self._state.window_close_event(event)

	def _evt_windowenter(self, event):
		self._state.window_enter(event)

	def _evt_windowexposed(self, event):
		self._state.window_exposed(event)

	def _evt_windowfocusgained(self, event):
		self._state.window_focus_gained(event)

	def _evt_windowfocuslost(self, event):
		self._state.window_focus_lost(event)

	def _evt_windowhidden(self, event):
		self._state.window_hidden(event)

	def _evt_windowhittest(self, event):
		self._state.window_hit_test(event)

	def _evt_windowleave(self, event):
		self._state.window_leave(event)

	def _evt_windowmaximized(self, event):
		self._state.window_maximized(event)

	def _evt_windowminimized(self, event):
		self._state.window_minimized(event)

	def _evt_windowmoved(self, event):
		self._state.window_moved(event)

	def _evt_windowresized(self, event):
		self._state.window_resized(event)

	def _evt_windowrestored(self, event):
		self._state.window_restored(event)

	def _evt_windowshown(self, event):
		self._state.window_shown(event)

	def _evt_windowsizechanged(self, event):
		self._state.window_size_changed(event)

	def _evt_windowtakefocus(self, event):
		self._state.window_take_focus(event)

	# Timers:

	def set_timeout(self, callback, milliseconds, **kwargs):
		"""
		Starts a timer using pygame.time.set_timer() which executes a given callback
		only once.

		The given "callback" is a function to execute after the "milliseconds"
		interval expires. Any keyword arguments after the "milliseconds"
		argument are passed as a dictionary to the given callback function.

		Returns an (integer) index identifying the timer, which can be used to cancel
		the timer by calling "clear_timeout()"
		"""
		return self.__set_timeout(callback, milliseconds, kwargs, False)

	def set_interval(self, callback, milliseconds, **kwargs):
		"""
		Starts a timer using pygame.time.set_timer() which executes a given callback
		at a repeated interval.

		The given "callback" is a function to execute after the "milliseconds"
		interval expires. Any keyword arguments after the "milliseconds"
		argument are passed as a dictionary to the given callback function.

		Returns an (integer) index identifying the timer, which can be used to cancel
		the timer by calling "clear_timeout()"
		"""
		return self.__set_timeout(callback, milliseconds, kwargs, True)

	def clear_timeout(self, timer_index):
		"""
		Clears a timer previously set using "set_timeout()"
		"""
		pygame.time.set_timer(USEREVENT + timer_index, 0)
		self.__timer_callbacks[timer_index] = None

	def __set_timeout(self, callback, milliseconds, arguments, recur):
		for timer_index in range(self.max_timers):
			if self.__timer_callbacks[timer_index] is None:
				self.__timer_callbacks[timer_index] = callback
				self.__timer_arguments[timer_index] = arguments
				self.__timer_recur_flag[timer_index] = recur
				pygame.time.set_timer(USEREVENT + timer_index, milliseconds)
				return timer_index
		raise RuntimeError("Too many timers!")

	def __timer_event(self, event):
		"""
		Called from the pygame event pump when a timer times out.
		Executes a timer event.
		"""
		timer_index = event.type - USEREVENT	# Subtract USEREVENT constant; indexes start with 0
		if self.__timer_callbacks[timer_index] is None:
			logging.warning("Timer event raised when corresponding callback not set")
		else:
			if len(self.__timer_arguments[timer_index]):
				self.__timer_callbacks[timer_index](self.__timer_arguments[timer_index])
			else:
				self.__timer_callbacks[timer_index]()
			if not self.__timer_recur_flag[timer_index]:
				self.clear_timeout(timer_index)

	def play(self, sound_name):
		"""
		Play a sound identified by "sound_name". If Game.quiet is True, does nothing.
		"""
		if not self.quiet:
			self.resources.sound(sound_name).play()


class GameState:

	def __init__(self, **kwargs):
		"""
		Set up this GameState to be the new game state next time through the main loop.
		The new state will have attributes set by keyword args passed to this function.
		If the current game state is an instance of "GameStateFinal", the current game
		state will not be changed.
		"""
		for varname, value in kwargs.items():
			setattr(self, varname, value)
		Game.current.change_state(self)

	def enter_state(self):
		"""
		Function called when the Game transitions to this state.
		Any information needed to be passed to this GameState should be passed as
		keyword args to the constructor.
		"""

	def exit_state(self, next_state):
		"""
		Function called when the Game transitions out of this state.
		The "next_state" parameter is the GameState object which will replace this one.
		"""

	# Early / late Game._main_loop() events:

	def loop_start(self):
		"""
		Called at the beginning of _main_loop() each time through, before processing events.
		The event loop looks like this:
		1. loop_start()                              <-- you are here
		2. event handling (keyboard, mouse, timers)
		3. loop_end()
		4. move the sprites
		5. update the display
		6. change to new game state (if needed)
		"""

	def loop_end(self):
		"""
		Called at the end of _main_loop() each time through.
		The event loop looks like this:
		1. loop_start()
		2. event handling (keyboard, mouse, timers)
		3. loop_end()                                <-- you are here
		4. move the sprites
		5. update the display
		6. change to new game state (if needed)
		"""

	# Event handlers called from Game._main_loop():

	def active_event(self, event):
		"""
		The "event" object has the following members:
			gain, state
		"""

	def audio_device_added(self, event):
		"""
		The "event" object has the following members:
			which, iscapture
		"""

	def audio_device_removed(self, event):
		"""
		The "event" object has the following members:
			which, iscapture
		"""

	def controller_axis_motion(self, event):
		"""
		The "event" object has the following members:
			[unknown]
		"""

	def controller_button_down(self, event):
		"""
		The "event" object has the following members:
			[unknown]
		"""

	def controller_button_up(self, event):
		"""
		The "event" object has the following members:
			[unknown]
		"""

	def controller_device_added(self, event):
		"""
		The "event" object has the following members:
			device_index
		"""

	def controller_device_remapped(self, event):
		"""
		The "event" object has the following members:
			instance_id
		"""

	def controller_device_removed(self, event):
		"""
		The "event" object has the following members:
			instance_id
		"""

	def drop_begin_event(self, event):
		"""
		The "event" object has the following members:
			none
		"""

	def drop_complete_event(self, event):
		"""
		The "event" object has the following members:
			none
		"""

	def drop_file_event(self, event):
		"""
		The "event" object has the following members:
			file
		"""

	def drop_text_event(self, event):
		"""
		The "event" object has the following members:
			text
		"""

	def finger_down(self, event):
		"""
		The "event" object has the following members:
			touch_id, finger_id, x, y, dx, dy
		"""

	def finger_motion(self, event):
		"""
		The "event" object has the following members:
			touch_id, finger_id, x, y, dx, dy
		"""

	def finger_up(self, event):
		"""
		The "event" object has the following members:
			touch_id, finger_id, x, y, dx, dy
		"""

	def joy_axis_motion(self, event):
		"""
		The "event" object has the following members:
			instance_id, axis, value
		"""

	def joy_ball_motion(self, event):
		"""
		The "event" object has the following members:
			instance_id, ball, rel
		"""

	def joy_button_down(self, event):
		"""
		The "event" object has the following members:
			instance_id, button
		"""

	def joy_button_up(self, event):
		"""
		The "event" object has the following members:
			instance_id, button
		"""

	def joy_device_added(self, event):
		"""
		The "event" object has the following members:
			device_index
		"""

	def joy_device_removed(self, event):
		"""
		The "event" object has the following members:
			instance_id
		"""

	def joy_hat_motion(self, event):
		"""
		The "event" object has the following members:
			instance_id, hat, value
		"""

	def key_down(self, event):
		"""
		The "event" object has the following members:
			key, mod, unicode, scancode
		"""

	def key_up(self, event):
		"""
		The "event" object has the following members:
			key, mod
		"""

	def midi_in(self, event):
		"""
		The "event" object has the following members:
			none
		"""

	def midi_out(self, event):
		"""
		The "event" object has the following members:
			none
		"""

	def mouse_button_down(self, event):
		"""
		The "event" object has the following members:
			pos, button
		"""

	def mouse_button_up(self, event):
		"""
		The "event" object has the following members:
			pos, button
		"""

	def mouse_motion(self, event):
		"""
		The "event" object has the following members:
			pos, rel, buttons
		"""

	def mouse_wheel(self, event):
		"""
		The "event" object has the following members:
			which, flipped, x, y
		"""

	def multi_gesture_event(self, event):
		"""
		The "event" object has the following members:
			touch_id, x, y, pinched, rotated, num_fingers
		"""

	def quit_event(self, event):
		"""
		The "event" object has the following members:
			none
		"""

	def sys_wm_event(self, event):
		"""
		The "event" object has the following members:
			[unknown]
		"""

	def text_editing_event(self, event):
		"""
		The "event" object has the following members:
			text, start, length
		"""

	def text_input_event(self, event):
		"""
		The "event" object has the following members:
			text
		"""

	def video_expose(self, event):
		"""
		The "event" object has the following members:
			none
		"""

	def video_resize(self, event):
		"""
		The "event" object has the following members:
			size, w, h
		"""

	def window_close_event(self, event):
		"""
		The "event" object has the following members:
			[unknown]
		"""

	def window_enter(self, event):
		"""
		The "event" object has the following members:
			[unknown]
		"""

	def window_exposed(self, event):
		"""
		The "event" object has the following members:
			[unknown]
		"""

	def window_focus_gained(self, event):
		"""
		The "event" object has the following members:
			[unknown]
		"""

	def window_focus_lost(self, event):
		"""
		The "event" object has the following members:
			[unknown]
		"""

	def window_hidden(self, event):
		"""
		The "event" object has the following members:
			[unknown]
		"""

	def window_hit_test(self, event):
		"""
		The "event" object has the following members:
			[unknown]
		"""

	def window_leave(self, event):
		"""
		The "event" object has the following members:
			[unknown]
		"""

	def window_maximized(self, event):
		"""
		The "event" object has the following members:
			[unknown]
		"""

	def window_minimized(self, event):
		"""
		The "event" object has the following members:
			[unknown]
		"""

	def window_moved(self, event):
		"""
		The "event" object has the following members:
			[unknown]
		"""

	def window_resized(self, event):
		"""
		The "event" object has the following members:
			[unknown]
		"""

	def window_restored(self, event):
		"""
		The "event" object has the following members:
			[unknown]
		"""

	def window_shown(self, event):
		"""
		The "event" object has the following members:
			[unknown]
		"""

	def window_size_changed(self, event):
		"""
		The "event" object has the following members:
			[unknown]
		"""

	def window_take_focus(self, event):
		"""
		The "event" object has the following members:
			[unknown]
		"""


class GameStateFinal(GameState):
	"""
	Final GameState; cannot be replaced even if a new GameState is instantiated
	after this one. See exit_states.py for an example.
	"""


#  end legame/game.py



#  legame/resources.py
#
#  Copyright 2020 - 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
"""
Provides the Resources class which loads images and sounds on demand, allowing
access to these based on keywords and indexes.
"""
import os, re
from pygame.image import load as image_load
from pygame.mixer import Sound


class Resources:

	def __init__(self, path = None, resource_dump = False):
		self.image_folder = os.path.join(path, "images")
		self.sounds_folder = os.path.join(path, "sounds")
		self.resource_dump = resource_dump
		self.sounds = {}
		self.images = {}
		self.image_sets = {}

	def sound(self, name):
		"""
		Return a "Sound" object from the file found at the given name.
		"""
		if name not in self.sounds:
			path = os.path.join(self.sounds_folder, name)
			if self.resource_dump:
				self.sounds[name] = path
			else:
				self.sounds[name] = Sound(path)
		return self.sounds[name]

	def image(self, name, **kwargs):
		"""
		Returns a pygame Surface with the given image loaded and converted to the
		screen format. You may provide a name which includes subpaths, i.e.:

			white_rook.image = <resources>.image("ChessPiece/White/rook.png")

		The actual filesystem path will be relative to the Resources "image_folder".

		"kwargs" may include: "convert", "convert_alpha", or "color_key". These trigger
		calling the appropriate pygame.Surface function after the image is loaded.

		From the pygame documentation for Surface.convert():

			[T]he new Surface will have the same pixel format as the display Surface. This
			is always the fastest format for blitting. It is a good idea to convert all
			Surfaces before they are blitted many times.

		Note: the "color_key" option overrides both "convert" and "convert_alpha".

		For a full description of alpha channels and color key alphas, see the pygame
		documentation.
		"""
		if name not in self.images:
			path = os.path.join(self.image_folder, name)
			if self.resource_dump:
				self.images[name] = path
			else:
				title, ext = os.path.splitext(name)
				self.images[name] = image_load(path)

				convert = kwargs.get("convert", False)
				convert_alpha = kwargs.get("convert_alpha", False)
				color_key = kwargs.get("color_key", None)

				if (convert or convert_alpha) and color_key is None:
					if convert_alpha:
						self.images[name].convert_alpha()
					else:
						self.images[name].convert()
				elif color_key is not None:
					self.images[name].set_colorkey(color_key)

		return self.images[name]

	def image_set(self, path, **kwargs):
		"""
		Returns an ImageSet, which provides a list of images and variants.
		For an explantation of the "convert_alpha" and "color_key" options, see
		Resource.image()
		"""
		if "/" in path:
			parts = path.split("/")
			name = parts[0]
			variants = parts[1:]
		else:
			name = path
			variants = []
		if name not in self.image_sets:
			self.image_sets[name] = ImageSet(self.image_folder, name,
				resource_dump = self.resource_dump, **kwargs)
		imgset = self.image_sets[name]
		for variant in variants: imgset = imgset.variants[variant]
		return imgset

	def preload_sounds(self):
		"""
		Loads all sound files found in Resources "sounds_folder".
		"""
		for filename in os.listdir(self.sounds_folder):
			self.sounds[filename] = Sound(os.path.join(self.sounds_folder, filename))

	def dump(self):
		"""
		Diagnostic function which dumps the filenames of the loaded resources to STDOUT.
		You must intialize the Resources using the "resource_dump" flag for the
		resources to load as path names rather than actual Image and Sound instances.
		"""
		if len(self.sounds):
			print("Sounds")
			print("-" * 40)
			for title in self.sounds: print(f"  {title}")
			print("")
		if len(self.images):
			print("Images")
			print("-" * 40)
			for title in self.images: print(f"  {title}")
			print("")
		print("Image Sets")
		print("-" * 40)
		for img_set in self.image_sets.values():
			img_set.dump()
		print("-" * 40)
		print("""
Note: Some resources might not be enumurated here.
If you use Flipper.preload(), only classes which subclass Flipper will have
image sets loaded. It may be possible to use an image set which is not cycled
using the Flipper class, in which case those images will have not been loaded.
			""")


class ImageSet:
	"""
	A class which keeps a group (or groups) of Image objects. ImageSets are used by
	the "Flipper" and "FlipEffect" classes to provide the images used by these
	classes to animate sprites. You can also use an ImageSet separately as well.

	An ImageSet contains a list of "images", which are pygame Surface objects
	created from loading image files. Additionally, the ImageSet has a dicitonary
	of "variants", which are themselves instances of ImageSet which contain images
	and variants. The "variants" correlate to subfolders. When the top-level
	ImageSet is loaded, all the variants are loaded as well.

	Images are referenced with the numeric indexes of "ImageSet.images". Variants
	are referenced with string keys of "ImageSet.variants".

	The parent ImageSet may or may not contain images. An image set which only
	contains variants which themselves contain images is perfectly okay. For
	example, you could have a top-level "ChessPiece" ImageSet which contains only
	two variants: "White" and "Black. These ImageSet objects would contain the
	actual images, one with white pieces, and one with black.

	This is best described using an illustration. Using the "ChessPiece" ImageSet
	example above, the folder hierarcy would look like this:

		<project folder>
		|-- <your_game.py>
		|-- resources
			|-- images
				|-- ChessPiece
					|-- Black
					|	|-- bishop.png
					|	|-- king.png
					|	|--	knight.png
					|	|-- pawn.png
					|	|-- queen.png
					|	|-- rook.png
					|-- White
						|-- bishop.png
						|-- king.png
						|--	knight.png
						|-- pawn.png
						|-- queen.png
						|-- rook.png

	When referencing the White pieces ImageSet, you would use the following
	syntax:

		white_pieces = ImageSet(<resources>.image_folder, "ChessPiece/White")

	...or...

		white_pieces = <resources>.image_set("ChessPiece/White")

	Note, however, that the individual images in the ImageSet instantiated above
	would still have numeric indexes, since the "images" attribute is a list, and
	not a dict. So you may not reference the individual white chess pieces by name,
	but will have to refer to them by their index, like so:

		white_rook.image = white_pieces.images[5]

	We can be sure that "rook" is the last item in the list with index "5" (out of
	6 pieces; list indexes always start at 0), because the ImageSet sorts the image
	files by name using a natural sort order. This is necessary for the the
	"Flipper" and "FlipEffect" classes.

	By "natural language sorting" we mean that numbers in the file name are sorted
	in the order that a human would expect. So "2" comes before "10". A simple sort
	of ASCII values doesn't do that:

		ASCII sort:
		-----------
		file10.png
		file15.png
		file1.png
		file20.png
		file25.png
		file2.png

		Nat sort:
		-----------
		file1.png
		file2.png
		file10.png
		file15.png
		file20.png
		file25.png

	If you would like to avoid using numeric indexes, consider using
	Resources.image() to load each image individually. In that case, the
	ChessPiece images might be referenced using the following syntax:

		white_rook.image = <resources>.image("ChessPiece/White/rook.png")

	"""

	image_extensions	= [".jpg", ".jpeg", ".png", ".gif", ".bmp", ".tga"]

	def __init__(self, images_dir, name, **kwargs):
		"""
		Creates an ImageSet by loading the images found in specified directory.
		"images_dir" is the directory that Resources uses as the base directory
		for images. This is commonly "<game directory>/resources/images", when using
		standard directory naming conventions.

		"name" will be the name that the ImagesSet is referred to as, and is also the
		name of the directory below "images_dir" where the image files may be found.

		"kwargs" may include: "convert", "convert_alpha", "color_key" or "resource_dump".

		When "resource_dump" is True, images and sounds will not be loaded. Instead,
		path names will be loaded in their place for debugging.

		For an explanation of "convert", "convert_alpha" and "color_key" options, see
		the Resource.image() function.
		"""

		self.variants = {}
		self.images = []
		self.name = name
		self.resource_dump = kwargs.get("resource_dump", False)
		convert = kwargs.get("convert", False)
		convert_alpha = kwargs.get("convert_alpha", True)
		color_key = kwargs.get("color_key", None)
		my_root = os.path.join(images_dir, name)
		if not os.path.isdir(my_root):
			raise NotADirectoryError(my_root)
		images = {}
		for entry in os.scandir(my_root):
			if entry.is_dir(follow_symlinks = True):
				self.variants[entry.name] = ImageSet(my_root, entry.name, **kwargs)
			elif entry.is_file(follow_symlinks = True):
				_, ext = os.path.splitext(entry.name)
				if ext in self.image_extensions:
					# append image to temporary dictionary, unsorted:
					if self.resource_dump:
						images[entry.name] = entry.path
					else:
						images[entry.name] = image_load(entry.path)
						if (convert or convert_alpha) and color_key is None:
							if convert_alpha:
								images[entry.name].convert_alpha()
							else:
								images[entry.name].convert()
						elif color_key is not None:
							images[entry.name].set_colorkey(color_key)
		self.count = len(images)
		self.last_index = self.count - 1
		if self.count:
			# Create list of images using "human" number sorting ("10" comes after "2")
			convert = lambda text: int(text) if text.isdigit() else text
			alphanum_key = lambda key: [ convert(c) for c in re.split('([0-9]+)', key) ]
			self.images = [images[key] for key in sorted(images, key = alphanum_key)]

	def variant(self, path):
		"""
		Returns an ImageSet which is a variant of this ImageSet.
		"""
		imgset = self
		for arg in path.split("/"):
			imgset = imgset.variants[arg]
		return imgset

	def dump(self):
		"""
		Print the path names of the images loaded for debugging.
		"""
		self._dump(self.name, 0)

	def _dump(self, keyname, indent):
		print(("   " * indent + "%s: %d images") % (keyname, self.count))
		for key, imgset in self.variants.items():
			imgset._dump(key, indent + 1)


#  end legame/resources.py



#!/usr/bin/python3
#  legame/templates/network-board-game.py
#
#  Copyright 2020 - 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
try:
	from pygame.locals import K_q, K_ESCAPE
except ImportError:
	from pygame import K_q, K_ESCAPE
from cable_car.json_messages import Message, MsgQuit
from legame.game import Game, GameState, GameStateFinal
from legame.board_game import Cell, GamePiece, GameBoard, BoardGame
from legame.network_game import NetworkGame
from legame.flipper import Flipper, FlipThrough, FlipNone


class MyGame(BoardGame, NetworkGame):

	def __init__ (self, options):
		self.set_resource_dir_from_file(__file__)
		BoardGame.__init__(self, options)
		NetworkGame.__init__(self, options)

	def run(self):
		return NetworkGame.run(self)

	def get_board(self):
		return MyBoard()

	def initial_state(self):
		return EmptyGameState()


class MyBoard(GameBoard):

	columns				= 8
	rows				= 8


class EmptyGameState(GameState):
	"""
	Example GameState subclass.
	Trim out whatever you don't need. All of these functions are included in the
	GameState class.
	"""

	def enter_state(self):
		"""
		Function called when the Game transitions TO this state.
		Any information needed to be passed to this GameState should be passed
		as keyword args to the constructor.
		"""

	def exit_state(self, next_state):
		"""
		Function called when the Game transitions OUT OF this state.
		The "next_state" parameter is the GameState object which will replace this one.
		"""

	def handle_message(self, message):
		"""
		Function called when a message comes over the wire through Game.current.messenger.
		"""
		if isinstance(message, MsgQuit):
			GSQuit(who = "them")

	def key_down(self, event):
		"""
		Key down event passed to this GameState.
		"event" will contain: key, mod, unicode, scancode
		"""
		if event.key in (K_ESCAPE, K_q):
			GSQuit(who = "me")

	def quit_event(self, event):
		"""
		Event handler called when the user clicks the window's close button.
		event will be empty
		"""
		GSQuit(who = "me")

	def key_up(self, event):
		"""
		Key up event passed to this GameState.
		"event" will contain: key, mod
		"""

	def mouse_motion(self, event):
		"""
		Mouse move event passed to this GameState.
		"event" will contain: pos, rel, buttons
		"""

	def mouse_button_down(self, event):
		"""
		Mouse down event passed to this GameState.
		"event" will contain: pos, button
		"""

	def mouse_button_up(self, event):
		"""
		Mouse up event passed to this GameState.
		"event" will contain: pos, button
		"""

	def active_event(self, event):
		"""
		"event" will contain: gain, state
		"""

	def joy_axis_motion(self, event):
		"""
		Joystick motion event passed to this GameState.
		"event" will contain: instance_id, axis, value
		"""

	def joy_ball_motion(self, event):
		"""
		Joystick ball motion event passed to this GameState.
		"event" will contain: instance_id, ball, rel
		"""

	def joy_hat_motion(self, event):
		"""
		Joystick hat motion event passed to this GameState.
		"event" will contain: instance_id, hat, value
		"""

	def joy_button_down(self, event):
		"""
		Joystick button down event passed to this GameState.
		"event" will contain: instance_id, button
		"""

	def joy_button_up(self, event):
		"""
		Joystick button up event passed to this GameState.
		"event" will contain: instance_id, button
		"""

	def video_resize(self, event):
		"""
		Event handler called when the window / display is resized.
		event will contain: size, w, h
		"""

	def video_expose(self, event):
		"""
		Event handler called when the window is exposed(?)
		event will be empty
		"""


class GSQuit(GameStateFinal):

	def enter_state(self):
		if self.who == "me":
			Game.current.messenger.send(MsgQuit())


class MsgAdd(Message):

	def encoded_attributes(self):
		return { "cell" : (self.cell.column, self.cell.row) }

	def decode_attributes(self, attributes):
		self.cell = Game.current.board.rotate(Cell(attributes["cell"][0], attributes["cell"][1]))


class MsgMove(Message):

	def encoded_attributes(self):
		return {
			"cell"			: (self.cell.column, self.cell.row),
			"target_cell"	: (self.target_cell.column, self.target_cell.row)
		}

	def decode_attributes(self, attributes):
		self.cell = Game.current.board.rotate(Cell(attributes["cell"][0], attributes["cell"][1]))
		self.target_cell = Game.current.board.rotate(
			Cell(attributes["target_cell"][0], attributes["target_cell"][1]))


class Piece(GamePiece, Flipper):

	def __init__(self, cell, color):
		self.color = color
		GamePiece.__init__(self, cell, color)
		self.image_folder = "Piece/" + color
		Flipper.__init__(self, FlipThrough("enter"), FlipNone())

	def update(self):
		GamePiece.update(self)
		Flipper.update(self)


if __name__ == '__main__':
	import argparse, sys, logging

	p = argparse.ArgumentParser()
	p.epilog = """
	Describe your game here.
	"""
	p.add_argument("--quiet", "-q", action = "store_true",
		help = "Don't make sound")
	p.add_argument("--verbose", "-v", action = "store_true",
		help = "Show more detailed debug information")
	p.add_argument("--resource-dump", "-r", action = "store_true",
		help = "Show sound and image resources for debugging")
	p.add_argument("--direct", "-d", action = "store_true",
		help = "Connect by ip address instead of using udp broadcast discovery.")
	options = p.parse_args()
	logging.basicConfig(
		level = logging.DEBUG if options.verbose else logging.ERROR,
		format = "[%(filename)24s:%(lineno)-4d] %(message)s"
	)

	game = MyGame(options)
	if options.resource_dump:
		Flipper.preload()
		game.resources.dump()
		sys.exit(0)
	else:
		sys.exit(game.run())


#  end legame/templates/network-board-game.py



#  legame/tests/flipper_test.py
#
#  Copyright 2020 - 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
import os
import pytest
import legame
from legame.game import Game
from legame.resources import Resources
from legame.flipper import Flipper, FlipThrough, FlipBetween

@pytest.fixture(autouse = True)
def thing():
	FakeGame()
	return Block()


class FakeGame:
	fps	= 30
	def __init__(self):
		Game.current = self
		examples = os.path.join(os.path.dirname(os.path.dirname(legame.__file__)), "examples")
		if not os.path.isdir(examples):
			raise NotADirectoryError(examples)
		self.resource_dir = os.path.join(os.path.abspath(examples), "resources")
		if not os.path.isdir(self.resource_dir):
			raise NotADirectoryError(self.resource_dir)
		self.resources = Resources(self.resource_dir, True)


class Block(Flipper):
	image_folder	= "Block/r"
	def __init__(self):
		Flipper.__init__(self, FlipThrough("enter"))

def test_flip_through_once(thing):
	assert isinstance(thing.flipper, FlipThrough)
	assert not thing.flipper.loop_forever
	assert thing.flipper.loops == 1
	assert thing.image == thing.flipper.image_set.images[0]
	for frame in [0, 1, 2]:
		assert thing.flipper.frame == frame
		thing.update()
	assert thing.flipper is None

def test_flip_through_twice(thing):
	thing.flip(FlipThrough("enter", loops = 2))
	assert not thing.flipper.loop_forever
	assert thing.flipper.loops == 2
	for frame in [0, 1, 2, 0, 1, 2]:
		assert thing.flipper.frame == frame
		thing.update()
	assert thing.flipper is None

def test_between_through_once(thing):
	thing.flip(FlipBetween("jiggle"))
	assert isinstance(thing.flipper, FlipBetween)
	assert not thing.flipper.loop_forever
	assert thing.flipper.loops == 1
	for frame in [0, 1, 2, 1, 0]:
		assert thing.flipper.frame == frame
		thing.update()
	assert thing.flipper is None

def test_flip_between_twice(thing):
	thing.flip(FlipBetween("jiggle", loops = 2))
	assert not thing.flipper.loop_forever
	assert thing.flipper.loops == 2
	assert thing.image == thing.flipper.image_set.images[0]
	for frame in [0, 1, 2, 1, 0, 1, 2, 1, 0]:
		assert thing.flipper.frame == frame
		thing.update()
	assert thing.flipper is None

def test_flip_through_loop_forever(thing):
	thing.flip(FlipThrough("enter", loop_forever = True))
	assert thing.flipper.loop_forever == True
	assert thing.flipper.frame == 0
	for frame in [0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2]:
		assert thing.flipper.frame == frame
		thing.update()

def test_two_flippers_queued(thing):
	thing.flip(FlipBetween("jiggle"), FlipThrough("enter"))
	assert thing.flipper.__class__.__name__ == "FlipBetween"
	assert not thing.flipper.loop_forever
	assert thing.flipper.loops == 1
	assert thing.flipper.frame == 0
	for frame in [0, 1, 2, 1, 0]:
		assert thing.flipper.frame == frame
		thing.update()
	assert thing.flipper.__class__.__name__ == "FlipThrough"
	assert not thing.flipper.loop_forever
	assert thing.flipper.loops == 1
	assert thing.flipper.frame == 0
	for frame in [0, 1, 2]:
		assert thing.flipper.frame == frame
		thing.update()
	assert thing.flipper is None

def test_starting_frame(thing):
	thing.flip(FlipBetween("jiggle", loop_forever = True, frame = 2, fps = 30))
	assert thing.flipper.loop_forever == True
	assert thing.flipper.frame == 2


#  end legame/tests/flipper_test.py



#  legame/__init__.py
#
#  Copyright 2020 - 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
"""
Modules which aid in creating 2-dimensional games with pygame.
"""
import math
from pygame.math import Vector2 as Vector

__version__ = "1.0.1"

PI						= math.pi
HALF_PI					= PI / 2.0
TWO_PI					= PI * 2.0
QUARTER_PI				= PI / 4.0

DEGREES_EAST			= 0.0
DEGREES_SOUTHEAST		= 45.0
DEGREES_SOUTH			= 90.0
DEGREES_SOUTHWEST		= 135.0
DEGREES_WEST			= 180.0
DEGREES_NORTHWEST		= -135.0
DEGREES_NORTH			= -90.0
DEGREES_NORTHEAST		= -45.0

RADIANS_EAST			= 0.0
RADIANS_SOUTHEAST		= QUARTER_PI
RADIANS_SOUTH			= HALF_PI
RADIANS_SOUTHWEST		= HALF_PI + QUARTER_PI
RADIANS_WEST			= PI
RADIANS_NORTHWEST		= -HALF_PI - QUARTER_PI
RADIANS_NORTH			= -HALF_PI
RADIANS_NORTHEAST		= -QUARTER_PI

RADIANS_ZERO			= RADIANS_EAST
RADIANS_45				= RADIANS_SOUTHEAST
RADIANS_90				= RADIANS_SOUTH
RADIANS_135				= RADIANS_SOUTHWEST
RADIANS_180				= RADIANS_WEST
RADIANS_NEG45			= RADIANS_NORTHEAST
RADIANS_NEG90			= RADIANS_NORTH
RADIANS_NEG135			= RADIANS_NORTHWEST

COMPASS_NORTH			= 0
COMPASS_NORTHEAST		= 1
COMPASS_EAST			= 2
COMPASS_SOUTHEAST		= 3
COMPASS_SOUTH			= 4
COMPASS_SOUTHWEST		= 5
COMPASS_WEST			= 6
COMPASS_NORTHWEST		= 7

COMPASS_DEGREES			= {
	COMPASS_NORTH: 270.0,
	COMPASS_NORTHEAST: 315.0,
	COMPASS_EAST: 0.0,
	COMPASS_SOUTHEAST: 45.0,
	COMPASS_SOUTH: 90.0,
	COMPASS_SOUTHWEST: 135.0,
	COMPASS_WEST: 180.0,
	COMPASS_NORTHWEST: 225.0
}

COMPASS_STR				= {
	COMPASS_NORTH: "N",
	COMPASS_NORTHEAST: "NE",
	COMPASS_EAST: "E",
	COMPASS_SOUTHEAST: "SE",
	COMPASS_SOUTH: "S",
	COMPASS_SOUTHWEST: "SW",
	COMPASS_WEST: "W",
	COMPASS_NORTHWEST: "NW"
}

SIDE_TOP				= COMPASS_NORTH
SIDE_RIGHT				= COMPASS_EAST
SIDE_BOTTOM				= COMPASS_SOUTH
SIDE_LEFT				= COMPASS_WEST

OFFSCREEN_TOP			= 0b1000
OFFSCREEN_BOTTOM		= 0b0100
OFFSCREEN_LEFT			= 0b0010
OFFSCREEN_RIGHT			= 0b0001
OFFSCREEN_TOPLEFT		= 0b1010
OFFSCREEN_TOPRIGHT		= 0b1001
OFFSCREEN_BOTTOMLEFT	= 0b0110
OFFSCREEN_BOTTOMRIGHT	= 0b0101

def deg2vector(degrees, magnitude = 1.0):
	vector = Vector()
	vector.from_polar((magnitude, degrees))
	return vector

def rad2vector(radians, magnitude = 1.0):
	vector = Vector()
	vector.from_polar((magnitude, math.degrees(radians)))
	return vector

def normal_radians(radians):
	"""
	Returns the radians given clamped to within the range -PI to PI
	"""
	radians = radians % TWO_PI
	if radians < -PI:
		radians += TWO_PI
	elif radians > PI:
		radians -= TWO_PI
	return radians

def turning_degrees(degrees):
	"""
	Clamps the given degrees to within range -180.0 to +180.0.
	"""
	if degrees < -180.0:
		while degrees < -180.0:
			degrees += 360.0
	elif degrees >= 180.0:
		while degrees >= 180.0:
			degrees -= 360.0
	return degrees

def normal_degrees(degrees):
	"""
	Clamps the given degrees to within range 0.0 to +359.999....
	"""
	return degrees % 360

def deg2side(degrees):
	"""
	Returns one of the "SIDE_<direction>" constants.
	"""
	degrees = normal_degrees(degrees)
	if degrees < 135.0:
		if degrees < 45.0:
			return SIDE_RIGHT
		return SIDE_BOTTOM
	if degrees < 225.0:
		return SIDE_LEFT
	if degrees > 315.0:
		return SIDE_RIGHT
	return SIDE_TOP

def rad2side(radians):
	"""
	Returns one of the "SIDE_<direction>" constants.
	"""
	radians = normal_radians(radians)
	if radians < RADIANS_NORTHEAST:
		if radians < RADIANS_NORTHWEST:
			return SIDE_LEFT
		return SIDE_TOP
	if radians > RADIANS_SOUTHEAST:
		if radians > RADIANS_SOUTHWEST:
			return SIDE_LEFT
		return SIDE_BOTTOM
	return SIDE_RIGHT

def deg2compass(degrees):
	"""
	Returns one of the "COMPASS_<direction>" constants.
	"""
	degrees = normal_degrees(degrees)
	if degrees < 157.0:
		if degrees < 67.0:
			if degrees < 22.0:
				return COMPASS_EAST
			return COMPASS_SOUTHEAST
		if degrees < 112.0:
			return COMPASS_SOUTH
		return COMPASS_SOUTHWEST
	if degrees < 247.0:
		if degrees < 202.0:
			return COMPASS_WEST
		return COMPASS_NORTHWEST
	if degrees < 337.0:
		if degrees < 292.0:
			return COMPASS_NORTH
		return COMPASS_NORTHEAST
	return COMPASS_EAST

def rad2compass(radians):
	"""
	Returns one of the "COMPASS_<direction>" constants.
	"""
	return deg2compass(math.degrees(radians))

def compass2deg(const):
	"""
	Returns the value in degrees of one of the "COMPASS_<direction>" constants.
	"""
	if const in COMPASS_DEGREES:
		return COMPASS_DEGREES[const]
	raise KeyError("Invalid value for compass constant")

def compass2rad(const):
	"""
	Returns the value in radians of one of the "COMPASS_<direction>" constants.
	"""
	return math.radians(compass2deg(const))

def compass_str(const):
	"""
	Returns a string, such as "N", "NW", for the given compass direction constant.
	"""
	if const in COMPASS_STR:
		return COMPASS_STR[const]
	raise KeyError("Invalid value for compass constant")

def triangular(value, reduction):
	"""
	Returns the number of times the "reduction" parameter must be applied to the
	"value" parameter for the value to reach zero.

	For example, "triangular(speed, deceleration)" will tell you how many frames it
	will take to decelerate from "speed" to zero, where "speed" and "deceleration"
	are measures of distance of travel in pixels per frame.
	"""
	return _inner_triangular(value + reduction, reduction)

def _inner_triangular(remainder, reduction):
	if remainder <= reduction:
		return reduction
	return remainder + _inner_triangular(remainder - reduction, reduction)

def vint(vector):
	"""
	Return a tuple (int) x, (int) y from a pygame.Vector2
	"""
	return int(vector.x), int(vector.y)

#  end legame/__init__.py



#  legame/tests/sprite_enhancement_test.py
#
#  Copyright 2020 - 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
from pygame import Rect
from legame.sprite_enhancement import MovingSprite, BoxedInSprite
from pygame.math import Vector2 as Vector
from legame import	normal_degrees, \
					DEGREES_EAST, DEGREES_NORTH, DEGREES_NORTHWEST, \
					DEGREES_WEST, DEGREES_SOUTHWEST, DEGREES_SOUTH, \
					COMPASS_WEST, COMPASS_NORTH, COMPASS_EAST, COMPASS_SOUTH


class Boxed(BoxedInSprite, MovingSprite):
	def __init__(self, box):
		MovingSprite.__init__(self)
		BoxedInSprite.__init__(self, box)


def test_constructor():
	thing = MovingSprite(9, 5)
	assert thing.x == 9
	assert thing.y == 5
	thing = MovingSprite(x = 9, y = 5, direction = 45, speed = 1000)
	assert thing.x == 9
	assert thing.y == 5
	assert thing.direction == 45
	assert thing.speed == 1000
	mag, deg = thing.motion.as_polar()
	assert deg == 45
	assert mag == 1000
	thing = MovingSprite(9, 5, 1000, 45)
	assert thing.x == 9
	assert thing.y == 5
	assert thing.direction == 45
	assert thing.speed == 1000

def test_motion():
	thing = MovingSprite(0, 0)
	thing.set_motion_polar(100, 90)
	assert thing.direction == 90
	assert thing.speed == 100

def test_basic_move():
	thing1 = MovingSprite(x = 100, y = 100)
	thing1.speed = 10
	thing1.direction = DEGREES_EAST
	thing1.cartesian_motion()
	assert thing1.x == 110
	assert thing1.y == 100
	thing1.x = 100
	assert thing1.x == 100
	assert thing1.direction == 0.0

def test_vector_copy():
	thing1 = MovingSprite(100, 100)
	thing2 = MovingSprite(0, 0)
	thing2.position = thing1.position
	assert thing2.x == thing1.x
	assert thing2.y == thing1.y

	thing1 = MovingSprite(80, 80)
	thing2 = MovingSprite()
	thing2.position = thing1.position
	assert thing2.x == thing1.x
	assert thing2.y == thing1.y

def test_boundary():

	thing = Boxed(Rect(0, 0, 40, 40))

	# Near the west wall

	thing.position = Vector(-5, 20)
	near = thing.nearest_boundary()
	assert near[0] == -5
	assert near[1] == COMPASS_WEST

	# Near the east wall
	thing.position = Vector(45, 20)
	near = thing.nearest_boundary()
	assert near[0] == -5
	assert near[1] == COMPASS_EAST

	# Near the north wall
	thing.position = Vector(20, -5)
	near = thing.nearest_boundary()
	assert near[0] == -5
	assert near[1] == COMPASS_NORTH

	# Near the south wall
	thing.position = Vector(20, 45)
	near = thing.nearest_boundary()
	assert near[0] == -5
	assert near[1] == COMPASS_SOUTH

def test_direction_to():
	tolerance = 0.0001
	thing = MovingSprite(0.0, 0.0)
	for degree in range(0, 720):
		target = MovingSprite(0, 0, degree, 10)
		target.cartesian_motion()
		dirc = normal_degrees(thing.direction_to_thing(target))
		ndeg = normal_degrees(degree)
		assert dirc - ndeg < tolerance \
			or dirc - 360 - ndeg < tolerance \
			or dirc + 360 - ndeg < tolerance

def test_turn_towards():

	tolerance = 0.0001

	# Setup subject in the middle of imaginary space
	subject = MovingSprite(x = 0.0, y = 0.0, direction = DEGREES_EAST, speed = 1.0)
	subject.max_turning_speed = 22.5

	# Setup target:
	target = MovingSprite(x = 0.0, y = 0.0)

	# Move target due south from center
	target.position = Vector(0.0, 0.0)
	target.direction = DEGREES_SOUTH
	target.cartesian_motion()

	# Check that subject turned south
	for stop in [22.5, 45.0, 67.5, 90.0]:
		subject.turn_towards(target.position)
		assert abs(subject.turning_speed - subject.max_turning_speed) < tolerance
		assert abs(subject.direction - stop) < tolerance

	# Reset the subject
	subject.direction = DEGREES_EAST
	subject.turning_speed = 0.0

	# Move target due north from center
	target.position = Vector(0.0, 0.0)
	target.direction = DEGREES_NORTH
	target.cartesian_motion()

	# Check that subject turned north
	for stop in [-22.5, -45.0, -67.5, -90.0]:
		subject.turn_towards(target.position)
		assert abs(subject.turning_speed + subject.max_turning_speed) < tolerance
		assert abs(subject.direction - stop) < tolerance

	# Reset the subject
	subject.direction = DEGREES_EAST
	subject.turning_speed = 0.0

	# Move target southwest from center
	target.position = Vector(0.0, 0.0)
	target.direction = DEGREES_SOUTHWEST
	target.cartesian_motion()

	# Check that subject turned south
	for stop in [22.5, 45.0, 67.5, 90.0]:
		subject.turn_towards(target.position)
		assert abs(subject.turning_speed - subject.max_turning_speed) < tolerance
		assert abs(subject.direction - stop) < tolerance

	# Reset the subject
	subject.direction = DEGREES_EAST
	subject.turning_speed = 0.0

	# Move target northwest from center
	target.position = Vector(0.0, 0.0)
	target.direction = DEGREES_NORTHWEST
	target.cartesian_motion()

	# Check that subject turned north
	for stop in [-22.5, -45.0, -67.5, -90.0]:
		subject.turn_towards(target.position)
		assert abs(subject.turning_speed + subject.max_turning_speed) < tolerance
		assert abs(subject.direction - stop) < tolerance

	# Reset the subject
	subject.direction = DEGREES_EAST
	subject.turning_speed = 0.0

	# Move target due west from center
	target.position = Vector(0.0, 0.0)
	target.direction = DEGREES_WEST
	target.cartesian_motion()

	# Check that subject turned north
	subject.turn_towards(target.position)
	subject.update()
	assert abs(subject.turning_speed) == subject.max_turning_speed
	assert abs(subject.direction) == abs(subject.direction)


#  end legame/tests/sprite_enhancement_test.py



#  legame/flipper.py
#
#  Copyright 2020 - 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
"""
Provides classes which are used to do image cycling on a Sprite.
"""
from collections import deque
from legame.game import Game


class Flipper:

	image_folder = None	# base name of this thing's image set. If none, use the class name

	@classmethod
	def preload(cls, **kwargs):
		for subclass in Flipper.__subclasses__():
			Game.current.resources.image_set(subclass.__name__, **kwargs)

	def __init__(self, *flippers, **kwargs):
		"""
		Initialize the image set used by this Thing, set the current FlipEffect to the first
		FlipEffect object given, and queue up any other FlipEffect given.

		e.g.:

			Flipper.__init__(self, <flipper>, <flipper>, <flipper>)

		See FlipEffect.__init__ for common image flipper options.
		"""
		self._base_image_set = Game.current.resources.image_set(
			self.__class__.__name__ if self.image_folder is None else self.image_folder,
			**kwargs
		)
		self._flipper_queue = deque()
		self.queue_flippers(flippers)
		self.next_flipper()

	def flip(self, *flippers):
		"""
		Replaces the current FlipEffect with the first FlipEffect object given, and queues any other
		FlipEffect objects given.

		e.g.:

			thing.flip(<flipper>, <flipper>, <flipper>)

		"""
		self._flipper_queue.clear()
		self.queue_flippers(flippers)
		self.next_flipper()

	def queue_flipper(self, flipper):
		"""
		Appends a single FlipEffect object to the queue.
		"""
		flipper.image_set = self._base_image_set if flipper.variant is None else self._base_image_set.variant(flipper.variant)
		self._flipper_queue.append(flipper)

	def queue_flippers(self, flippers):
		"""
		Appends a list of FlipEffect objects to the queue.
		"""
		for flipper in flippers:
			flipper.image_set = self._base_image_set if flipper.variant is None else self._base_image_set.variant(flipper.variant)
		self._flipper_queue.extend(flippers)

	def update(self):
		"""
		Called from pygame.Sprite, this function updates the "image" property of the sprite.
		If your Sprite uses function as the primary "update" function without modifying it,
		make sure that this function is first in your Sprite's method resolution order (MRO).
		"""
		if self.flipper is None:
			return
		self.image = self.flipper.update()
		if self.flipper.done:
			self.next_flipper()

	def next_flipper(self):
		"""
		Advances to the next FlipEffect in the queue.
		"""
		if len(self._flipper_queue):
			self.flipper = self._flipper_queue.popleft()
			self.image = self.flipper.first_image()
		else:
			self.flipper = None


class FlipEffect:
	"""
	Sets the image on a Sprite to a member of an ImageSet in sequence
	"""

	def __init__(self, variant = None, on_complete = None, **kwargs):
		"""
		When an FlipEffect is queued in a Flipper, it gets a reference to the root of the
		Flipper's ImageSet. Set the variant of the ImageSet to flip using the "variant"
		parameter (string).

		If you would an action to occur when the sequence is complete, pass a function
		to the "on_complete" parameter.

		**kwargs may include:

			loop_forever:	Restart from the beginning after finishing the cycle

			loops:			Number of loops to cycle through, when not "loop_forever"

			fps:			Frames-per-second. The FlipEffect will use the game.fps
							to determine how many frames to skip between flips.

			frame:			The frame to start at

		These vary according to the particular FlipEffect subclass you're using. The
		ones defined in this module include:

			FlipThrough, FlipBetween, FlipNone

		"""
		self.variant = variant
		self.on_complete_function = on_complete
		self.loop_forever = kwargs.pop("loop_forever", False)
		self.loops = kwargs.pop("loops", 1)
		self.fps = kwargs.pop("fps", None)
		self.frame = kwargs.pop("frame", 0)
		if len(kwargs): logging.warning("Unexpected keyword arguments: %s" % kwargs)
		self.__updates_per_frame = 1 if self.fps is None \
			else Game.current.fps // self.fps if Game.current is not None \
			else 60 // self.fps
		self._updates_this_frame = 0
		self._loops_remaining = self.loops
		self.done = False

	def first_image(self):
		"""
		Returns the first image in the set, or the only image in the case of FlipNone.
		"""
		return self.image_set.images[self.frame]

	def update(self):
		"""
		Called from Flipper.update() - advance frame in accordance with fps.
		Returns Image object.
		"""
		if not self.done:
			self._updates_this_frame += 1
			if self._updates_this_frame >= self.__updates_per_frame:
				self._updates_this_frame = 0
				self.advance_frame()
				if self.done and self.on_complete_function:
					self.on_complete_function()
		return self.image_set.images[self.frame]


class FlipNone(FlipEffect):
	"""
	Displays only the first image in the ImageSet
	"""

	def first_image(self):
		self.done = True
		return FlipEffect.first_image(self)

	def advance_frame(self):
		pass


class FlipThrough(FlipEffect):
	"""
	Cycles from the start to the end of an ImageSet. If looping, starts again from the beginning,
	otherwise leaves the frame pointer at the end of the image set.
	"""

	def advance_frame(self):
		if self.frame == self.image_set.last_index:
			if self.loop_forever:
				self.frame = 0
			else:
				self._loops_remaining -= 1
				if self._loops_remaining:
					self.frame = 0
				else:
					self.done = True
		else:
			self.frame += 1


class FlipBetween(FlipEffect):
	"""
	Cycles back and forth through an ImageSet; when at the end, backs up to the beginning
	"""

	def __init__(self, variant = None, **kwargs):
		FlipEffect.__init__(self, variant, **kwargs)
		self.__direction = 1

	def advance_frame(self):
		if self.frame == self.image_set.last_index:
			self.__direction = -1
		elif self.frame == 0 and self.__direction < 0:
			self.__direction = 1
			if not self.loop_forever:
				self._loops_remaining -= 1
				self.done = self._loops_remaining == 0
				if self.done:
					return
		self.frame += self.__direction


#  end legame/flipper.py



#  legame/callout.py
#
#  Copyright 2020 - 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
"""
Provides the Callout class, a Sprite used during development to provide debug
information positioned near another animated sprite.
"""
from pygame import Rect, Surface
from pygame.sprite import Sprite


class Callout(Sprite):

	def __init__(self, sprite, group, font):
		Sprite.__init__(self, group)
		self.sprite = sprite
		self.rect = Rect((sprite.rect.right, sprite.rect.bottom, 0, 0))
		self.image = Surface((0, 0))
		self.font = font
		self.empty()

	def empty(self):
		self.texts = []
		self.rect.width = 0

	def write(self, text, color = (255, 255, 255)):
		self.texts.append(self.font.render(text, True, color))
		width, height = self.font.size(text)
		if width > self.rect.width:
			self.rect.width = width
		self.rect.height += height

	def update(self):
		self.rect.left = self.sprite.rect.right
		self.rect.top = self.sprite.rect.bottom
		self.image = Surface((self.rect.width, self.rect.height))
		self.image.set_colorkey((0,0,0))
		self.image.fill((0,0,0))
		line_height = self.font.get_linesize()
		y = 0
		for s in self.texts:
			self.image.blit(s, (0, y))
			y += line_height
		self.rect.height = y


#  end legame/callout.py



#  legame/tests/board_game_test.py
#
#  Copyright 2020 - 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
import pytest
from pygame import Rect
from legame.board_game import BoardGame, GameBoard, AbstractGamePiece, Cell
from legame.game import Game

@pytest.fixture(autouse = True)
def game():
	return FakeGame()


class FakeGame(BoardGame):
	pass


def test_game_init():
	FakeGame()
	assert isinstance(Game.current, BoardGame)
	assert isinstance(Game.current.board, GameBoard)
	assert Game.current.board.left == 0
	assert Game.current.board.top == 0
	assert Game.current.board.cell_height == 50
	assert Game.current.board.cell_width == 50
	assert Game.current.board.cell_half_width == 25
	assert Game.current.board.cell_half_height == 25

def test_cell_str():
	game = FakeGame()
	board = game.board
	cell = Cell(5, 6)
	assert str(cell) == "Cell at column 5, row 6"

def test_cell_at():
	game = FakeGame()
	board = game.board

	with pytest.raises(ValueError) as e:
		x = board.cell_at()
	with pytest.raises(ValueError) as e:
		x = board.cell_at(())
	with pytest.raises(ValueError) as e:
		x = board.cell_at(1)
	with pytest.raises(ValueError) as e:
		x = board.cell_at((1,))
	with pytest.raises(ValueError) as e:
		x = board.cell_at(1,2,3)
	with pytest.raises(ValueError) as e:
		x = board.cell_at((1,2,3))

	cell = board.cell_at(0, 0)
	assert isinstance(cell, Cell)
	assert cell.column == 0
	assert cell.row == 0
	cell = board.cell_at((0, 0))
	assert isinstance(cell, Cell)
	assert cell.column == 0
	assert cell.row == 0

	# floats:

	cell = board.cell_at(25.0, 25.0)
	assert isinstance(cell, Cell)
	assert cell.column == 0
	assert cell.row == 0
	cell = board.cell_at((25.0, 25.0))
	assert isinstance(cell, Cell)
	assert cell.column == 0
	assert cell.row == 0

	cell = board.cell_at((1, 1))
	assert isinstance(cell, Cell)
	assert cell.column == 0
	assert cell.row == 0

	cell = board.cell_at(25, 25)
	assert isinstance(cell, Cell)
	assert cell.column == 0
	assert cell.row == 0

	cell = board.cell_at((26, 26))
	assert isinstance(cell, Cell)
	assert cell.column == 0
	assert cell.row == 0

	cell = board.cell_at((49, 49))
	assert isinstance(cell, Cell)
	assert cell.column == 0
	assert cell.row == 0

	cell = board.cell_at((50, 50))
	assert isinstance(cell, Cell)
	assert cell.column == 1
	assert cell.row == 1

	cell = board.cell_at((99, 99))
	assert isinstance(cell, Cell)
	assert cell.column == 1
	assert cell.row == 1

	cell = board.cell_at((100, 100))
	assert isinstance(cell, Cell)
	assert cell.column == 2
	assert cell.row == 2

	# test outside board

	cell = board.cell_at(-1, 0)
	assert cell is None

	cell = board.cell_at(0, -1)
	assert cell is None

	cell = board.cell_at(board.rect.width + 1, 0)
	assert cell is None

	cell = board.cell_at(0, board.rect.height + 1)
	assert cell is None

def test_cell_get_rect():
	game = FakeGame()
	board = game.board

	cell = board.cell_at(0, 0)
	assert isinstance(cell, Cell)
	assert cell.column == 0
	assert cell.row == 0
	rect = cell.rect()
	assert isinstance(rect, Rect)
	assert rect.x == 0
	assert rect.y == 0
	assert rect.width == game.board.cell_width
	assert rect.height == game.board.cell_height

	cell = board.cell_at(75, 75)
	assert isinstance(cell, Cell)
	assert cell.column == 1
	assert cell.row == 1
	rect = cell.rect()
	assert isinstance(rect, Rect)
	assert rect.x == game.board.cell_width
	assert rect.y == game.board.cell_height
	assert rect.width == game.board.cell_width
	assert rect.height == game.board.cell_height

def test_cell_rotate():
	game = FakeGame()
	board = game.board

	cell = Cell(0, 0)
	rot_cell = board.rotate(cell)
	assert rot_cell.column + cell.column == board.last_column
	assert rot_cell.row + cell.row == board.last_row

	cell = Cell(2, 4)
	rot_cell = board.rotate(cell)
	print(cell, rot_cell)
	assert rot_cell.column + cell.column == board.last_column
	assert rot_cell.row + cell.row == board.last_row

	cell = Cell(board.center_column, board.center_row)
	rot_cell = board.rotate(cell)
	assert rot_cell.column + cell.column == board.last_column
	assert rot_cell.row + cell.row == board.last_row

def test_cell_access():
	game = FakeGame()
	board = game.board

	cell = Cell(0, 0)
	piece = board.piece_at(cell)
	assert piece is None
	board.set_cell(cell, AbstractGamePiece(cell, "r"))
	piece = board.piece_at(cell)
	assert isinstance(piece, AbstractGamePiece)
	assert piece.color == "r"

	cell = Cell(0, 1)
	piece = board.piece_at(cell)
	assert piece is None
	board.set_cell(cell, AbstractGamePiece(cell, "r"))
	piece = board.piece_at(cell)
	assert isinstance(piece, AbstractGamePiece)
	assert piece.color == "r"

	cell = Cell(0, 0)
	board.clear_cell(cell)
	piece = board.piece_at(cell)
	assert piece is None

	cell = Cell(1, 0)
	piece = cell.piece()
	assert piece is None
	assert cell.is_empty()
	cell.set(AbstractGamePiece(cell, "r"))
	piece = cell.piece()
	assert isinstance(piece, AbstractGamePiece)
	assert piece.color == "r"

	cell = Cell(1, 1)
	piece = cell.piece()
	assert piece is None
	cell.set(AbstractGamePiece(cell, "r"))
	piece = cell.piece()
	assert isinstance(piece, AbstractGamePiece)
	assert piece.color == "r"
	game.my_color = piece.color
	assert cell.is_mine()

	cell = Cell(0, 0)
	cell.clear()
	piece = board.piece_at(cell)
	assert piece is None
	assert not cell.is_mine()

def test_column_row_access():
	game = FakeGame()
	board = game.board

	cells = board.column(0)
	assert isinstance(cells, list)
	assert len(cells) == board.rows

	cells = board.row(0)
	assert isinstance(cells, list)
	assert len(cells) == board.columns

	cell = Cell(0, 0)
	board.set_cell(cell, AbstractGamePiece(cell, "r"))
	cells = board.row(cell.row)
	assert isinstance(cells[0], AbstractGamePiece)
	for cell in cells[1:]: assert cell is None

	cell = Cell(0, 0)
	board.set_cell(cell, AbstractGamePiece(cell, "r"))
	cells = board.column(cell.column)
	assert isinstance(cells[0], AbstractGamePiece)
	for cell in cells[1:]: assert cell is None

	cell = Cell(board.last_column, board.last_row)
	board.set_cell(cell, AbstractGamePiece(cell, "r"))
	cells = board.row(cell.row)
	assert isinstance(cells[-1], AbstractGamePiece)
	for cell in cells[:-1]: assert cell is None

	cell = Cell(board.last_column, board.last_row)
	board.set_cell(cell, AbstractGamePiece(cell, "r"))
	cells = board.column(cell.column)
	assert isinstance(cells[-1], AbstractGamePiece)
	for cell in cells[:-1]: assert cell is None

def test_cell_copy():
	game = FakeGame()
	board = game.board
	cell1 = Cell(5, 6)
	cell2 = cell1.copy()
	assert cell2.column == cell1.column
	assert cell2.row == cell1.row

	cell2 = cell1.shifted(1)
	assert cell2.column == cell1.column + 1
	assert cell2.row == cell1.row

	cell2 = cell1.shifted(0, 1)
	assert cell2.column == cell1.column
	assert cell2.row == cell1.row + 1

	cell2 = cell1.shifted(-1)
	assert cell2.column == cell1.column - 1
	assert cell2.row == cell1.row

	cell2 = cell1.shifted(0, -1)
	assert cell2.column == cell1.column
	assert cell2.row == cell1.row - 1

	cell2 = cell1.moved(row = 7)
	assert cell2.column == cell1.column
	assert cell2.row == 7

	cell2 = cell1.moved(column = 5)
	assert cell2.column == 5
	assert cell2.row == cell1.row

def test_cell_eq():
	game = FakeGame()
	board = game.board

	cell1 = Cell(5, 6)
	cell2 = cell1.copy()
	assert cell1 == cell2

def test_cell_unpacking():
	game = FakeGame()
	board = game.board
	cell = Cell(5, 6)
	column, row = cell
	assert column == 5
	assert row == 6

def test_game_piece_travel():
	game = FakeGame()
	board = game.board
	piece = AbstractGamePiece(Cell(5,6), "r")
	piece.move_to(Cell(3,4))
	assert piece.target_cell.column == 3
	assert piece.target_cell.row == 4
	piece.move_to(column = 3)
	assert piece.target_cell.column == 3
	assert piece.target_cell.row == 6
	piece.move_to(row = 1)
	assert piece.target_cell.column == 5
	assert piece.target_cell.row == 1
	piece.move_to(columns = -1)
	assert piece.target_cell.column == 4
	assert piece.target_cell.row == 6
	piece.move_to(rows = -2)
	assert piece.target_cell.column == 5
	assert piece.target_cell.row == 4


if __name__ == "__main__":
	test_game_piece_travel()


#  end legame/tests/board_game_test.py



#  legame/network_game.py
#
#  Copyright 2020 - 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
"""
Provides the NetworkGame class, a framework for games played over a network.
"""
import importlib
from time import time
from legame.game import Game, GameState
from legame.joiner import BroadcastJoiner, DirectJoiner


class NetworkGame(Game):
	"""
	Basic framework for a game to be played over a network. Requires: cable_car.
	.
	There are two basic methods to use to connect to other computers on the
	network. You can use "broadcast connections", which announce your availability
	using UDP broadcast, or by specifying either "client" or "server" mode in the
	given "options". If neither "client" or "server" is given, broadcast is used
	by default.

	Message transport selection is up to you. The current options are "json" and "byte".
	See the cable_car docs for more info on message transports.
	"""

	udp_port		= 8222		# Port to broadcast on
	tcp_port		= 8223		# Port to listen on
	direct			= False		# Connect directly, instead of using udp broadcast
	transport		= "json"	# cable_car transport to use.
	xfer_interval	= 0.125		# Number of seconds between calls to service the messenger
	connect_timeout	= 10.0		# Number of seconds to wait before giving up when connecting

	def __init__(self, options = None):
		"""
		Network game constructor.
		.

		The "options" argument is expected to be a dictionary, the items of which are
		set as attributes of the game during initialization. Some appropriate key/value
		pairs to pass to the __init__ function would be:

			direct
			tcp_port
			udp_port
			transport
			xfer_interval

		... in addition to the common Game class options:

			display_depth
			fps
			fullscreen
			quiet
			resource_dump

		"""
		if options is not None:
			for varname, value in options.__dict__.items():
				setattr(self, varname, value)
		module = importlib.import_module("cable_car.%s_messages" % self.transport)
		globals().update({ name: module.__dict__[name] for name in module.__dict__})
		if self.direct:
			self.__joiner = DirectJoiner(options)
		else:
			self.__joiner = BroadcastJoiner(options)

	def run(self):
		self.__joiner.show()
		if not self.__joiner.messenger: return 5
		self.messenger = self.__joiner.messenger
		del self.__joiner
		self._next_xfer = time()
		try:
			return Game.run(self)
		except Exception as e:
			self.messenger.send(MsgQuit())
			self.messenger.shutdown()
			return 1

	def _end_loop(self):
		"""
		Called at the end of the _main_loop(), this function handles message transfer.
		"""
		self._state.loop_end()
		if time() >= self._next_xfer:
			self._next_xfer = time() + self.xfer_interval
			self.messenger.xfer()
			message = self.messenger.get()
			while message is not None:
				self._state.handle_message(message)
				message = self.messenger.get()

# Dynamically append "handle_message" method used by NetworkGame to the GameState class:

def _handle_message(self, message):
	pass


GameState.handle_message = _handle_message


#  end legame/network_game.py



#  legame/tests/functions_test.py
#
#  Copyright 2020 - 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
from math import radians
from legame import deg2side, rad2side, normal_radians, normal_degrees, \
	SIDE_LEFT, SIDE_TOP, SIDE_RIGHT, SIDE_BOTTOM, PI, TWO_PI

def test_normal_degrees():
	for degrees in range(-3600, 3600, 45):
		assert normal_degrees(degrees) >= 0
		assert normal_degrees(degrees) <= 360

def test_normal_radians():
	rad = -TWO_PI * 3
	while rad <= TWO_PI * 3:
		assert normal_radians(rad) <= PI
		assert normal_radians(rad) >= -PI
		rad += TWO_PI / 8

def test_deg2side():
	assert deg2side(0) == SIDE_RIGHT
	assert deg2side(90) == SIDE_BOTTOM
	assert deg2side(180) == SIDE_LEFT
	assert deg2side(270) == SIDE_TOP
	assert deg2side(44) == SIDE_RIGHT
	assert deg2side(46) == SIDE_BOTTOM
	assert deg2side(134) == SIDE_BOTTOM
	assert deg2side(136) == SIDE_LEFT
	assert deg2side(224) == SIDE_LEFT
	assert deg2side(226) == SIDE_TOP
	assert deg2side(314) == SIDE_TOP
	assert deg2side(316) == SIDE_RIGHT

def test_rad2side():
	assert rad2side(radians(0)) == SIDE_RIGHT
	assert rad2side(radians(90)) == SIDE_BOTTOM
	assert rad2side(radians(180)) == SIDE_LEFT
	assert rad2side(radians(270)) == SIDE_TOP
	assert rad2side(radians(44)) == SIDE_RIGHT
	assert rad2side(radians(46)) == SIDE_BOTTOM
	assert rad2side(radians(134)) == SIDE_BOTTOM
	assert rad2side(radians(136)) == SIDE_LEFT
	assert rad2side(radians(224)) == SIDE_LEFT
	assert rad2side(radians(226)) == SIDE_TOP
	assert rad2side(radians(314)) == SIDE_TOP
	assert rad2side(radians(316)) == SIDE_RIGHT


#  end legame/tests/functions_test.py
